""" The definition of the VLR Header, VLR, the KnownVLRs
 are in this module.

 A KnownVLR is a VLR for which we know how to parse its record_data
"""
import abc
import ctypes
import logging
import struct

from .rawvlr import NULL_BYTE, BaseVLR, VLR
from .. import extradims
from ..extradims import (
    get_type_for_extra_dim,
    get_kind_of_extra_dim,
)
from ..point.dims import DimensionInfo, DimensionKind

abstractmethod = abc.abstractmethod

logger = logging.getLogger(__name__)


class IKnownVLR(abc.ABC):
    """Interface that any KnownVLR must implement.
    A KnownVLR is a VLR for which we know how to parse its record_data

    Implementing this interfaces allows to automatically call the
    right parser for the right VLR when reading them.
    """

    @staticmethod
    @abstractmethod
    def official_user_id():
        """Shall return the official user_id as described in the documentation"""
        pass

    @staticmethod
    @abstractmethod
    def official_record_ids():
        """Shall return the official record_id for the VLR

        .. note::

            Even if the VLR has one record_id, the return type must be a tuple

        Returns
        -------
        tuple of int
            The record_ids this VLR type can have
        """
        pass

    @abstractmethod
    def record_data_bytes(self):
        """Shall return the bytes corresponding to the record_data part of the VLR
        as they should be written in the file.

        Returns
        -------
        bytes
            The bytes of the vlr's record_data

        """
        pass

    @abstractmethod
    def parse_record_data(self, record_data):
        """Shall parse the given record_data into a user-friendlier structure

        Parameters
        ----------
        record_data: bytes
            The record_data bytes read from the file

        """
        pass


class BaseKnownVLR(BaseVLR, IKnownVLR):
    """Base Class to factorize common code between the different type of Known VLRs"""

    def __init__(self, record_id=None, description=""):
        super().__init__(
            self.official_user_id(),
            self.official_record_ids()[0] if record_id is None else record_id,
            description,
        )

    @classmethod
    def from_raw(cls, raw):
        vlr = cls()
        vlr.description = raw.header.description.decode("ascii")
        vlr.parse_record_data(raw.record_data)
        return vlr


class ClassificationLookupVlr(BaseKnownVLR):
    """This vlr maps class numbers to short descriptions / names

    >>> lookup = ClassificationLookupVlr()
    >>> lookup[0] = "never_classified"
    >>> lookup[2] = "ground"
    >>> lookup[0]
    'never_classified'
    """

    _lookup_struct = struct.Struct("<B15s")

    def __init__(self):
        super().__init__(description="Classification Lookup")
        self.lookups = {}

    def parse_record_data(self, record_data):
        for class_id, desc in struct.iter_unpack("<B15s", record_data):
            # index using desc[i:i+1], because desc[i] gives an int, and we want a byte
            description = b"".join(
                desc[i:i+1] for i in range(len(desc)) if desc[i:i+1].isalnum() or desc[i:i+1] == b' '
            ).decode()
            self.lookups[class_id] = description

    def record_data_bytes(self):
        def lookup_converter(lookup_dict):
            for class_id, description in lookup_dict.items():
                description_bytes = description.encode("ascii")
                if len(description_bytes) > 15:
                    raise ValueError(
                        "decription ({}) is to long ({} bytes), it must not exceed 15 bytes when encoded".format(
                            description, len(description_bytes)
                        )
                    )
                yield class_id, description_bytes

        return b"".join(
            self._lookup_struct.pack(class_id, desc)
            for class_id, desc in lookup_converter(self.lookups)
        )

    def __getitem__(self, class_id):
        return self.lookups[class_id]

    def __setitem__(self, class_id, description):
        if class_id not in range(256):
            raise ValueError("Class id {} is not in range [0, 255]".format(class_id))

        self.lookups[class_id] = description

    @staticmethod
    def official_user_id():
        return "LASF_Spec"

    @staticmethod
    def official_record_ids():
        return (0,)


class LasZipVlr(BaseKnownVLR):
    """Contains the information needed by laszip (or any other laz backend)
    to compress the point records.
    """

    def __init__(self, data):
        super().__init__(description="http://laszip.org")
        self.record_data = data

    def parse_record_data(self, record_data):
        # Only laz backends know how to parse this
        pass

    def record_data_bytes(self):
        return self.record_data

    @staticmethod
    def official_user_id():
        return "laszip encoded"

    @staticmethod
    def official_record_ids():
        return (22204,)

    @classmethod
    def from_raw(cls, raw_vlr):
        return cls(raw_vlr.record_data)


class ExtraBytesStruct(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("reserved", ctypes.c_uint8 * 2),
        ("data_type", ctypes.c_uint8),
        ("options", ctypes.c_uint8),
        ("name", ctypes.c_char * 32),
        ("unused", ctypes.c_uint8 * 4),
        ("_no_data", (ctypes.c_byte * 8) * 3),
        ("_min", (ctypes.c_byte * 8) * 3),
        ("_max", (ctypes.c_byte * 8) * 3),
        ("_scale", (ctypes.c_byte * 8) * 3),
        ("_offset", (ctypes.c_byte * 8) * 3),
        ("description", ctypes.c_char * 32),
    ]

    _uint64t_struct = struct.Struct("<Q")
    _int64t_struct = struct.Struct("<q")
    _double_struct = struct.Struct("<d")

    def _struct_parser_for_kind(self):
        signedness = get_kind_of_extra_dim(self.data_type)

        if signedness == DimensionKind.FloatingPoint:
            return self._double_struct
        elif signedness == DimensionKind.SignedInteger:
            return self._int64t_struct
        elif signedness == DimensionKind.UnsignedInteger:
            return self._uint64t_struct
        else:
            return None

    def _parse_special_property(self, name):
        strct = self._struct_parser_for_kind()
        return tuple(strct.unpack(d)[0] for d in getattr(self, name))

    @classmethod
    def from_dimension_info(cls, info: DimensionInfo) -> "ExtraBytesStruct":
        if info.is_standard:
            raise ValueError("ExtraBytesStruct cannot describe standard dims")

        if info.kind == DimensionKind.BitField:
            raise ValueError("ExtraBytesStruct cannot describe bit fields")

        type_str = info.type_str()
        assert type_str is not None
        if type_str.endswith("u1"):
            extra_byte = cls(
                data_type=0,
                name=info.name.encode(),
                description="".encode(),
                options=int(type_str[:-2]),
            )
        else:
            if info.num_elements > 3:
                raise ValueError("Only u1 data type supports more than 3 elements")
            type_id = extradims.get_id_for_extra_dim_type(type_str)
            extra_byte = cls(
                data_type=type_id,
                name=info.name.encode(),
                description="".encode(),
            )

        return extra_byte

    @property
    def no_data(self):
        return self._parse_special_property("_no_data")

    @property
    def min(self):
        return self._parse_special_property("_min")

    @property
    def max(self):
        return self._parse_special_property("_max")

    @property
    def offset(self):
        return self._parse_special_property("_offset")

    @property
    def scale(self):
        return self._parse_special_property("_scale")

    def format_name(self):
        return self.name.rstrip(NULL_BYTE).decode().replace(" ", "_").replace("-", "_")

    def type_tuple(self):
        if self.data_type == 0:
            return self.format_name(), "{}u1".format(self.options)
        return self.format_name(), get_type_for_extra_dim(self.data_type)

    @staticmethod
    def size():
        return ctypes.sizeof(ExtraBytesStruct)

    def __repr__(self):
        return "<ExtraBytesStruct({}, {}, {})>".format(
            *self.type_tuple(), self.description
        )


class ExtraBytesVlr(BaseKnownVLR):
    def __init__(self):
        super().__init__(description="Extra Bytes Record")
        self.extra_bytes_structs = []

    def parse_record_data(self, data):
        if (len(data) % ExtraBytesStruct.size()) != 0:
            raise ValueError(
                "Data length of ExtraBytes vlr must be a multiple of {}".format(
                    ExtraBytesStruct.size()
                )
            )
        num_extra_bytes_structs = len(data) // ExtraBytesStruct.size()
        self.extra_bytes_structs = [None] * num_extra_bytes_structs
        for i in range(num_extra_bytes_structs):
            self.extra_bytes_structs[i] = ExtraBytesStruct.from_buffer_copy(
                data[ExtraBytesStruct.size() * i : ExtraBytesStruct.size() * (i + 1)]
            )

    def record_data_bytes(self):
        return b"".join(
            bytes(extra_struct) for extra_struct in self.extra_bytes_structs
        )

    def type_of_extra_dims(self):
        return [extra_dim.type_tuple() for extra_dim in self.extra_bytes_structs]

    def __repr__(self):
        return "<ExtraBytesVlr(extra bytes structs: {})>".format(
            len(self.extra_bytes_structs)
        )

    @staticmethod
    def official_user_id():
        return "LASF_Spec"

    @staticmethod
    def official_record_ids():
        return (4,)


class WaveformPacketStruct(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("bits_per_sample", ctypes.c_ubyte),
        ("waveform_compression_type", ctypes.c_ubyte),
        ("number_of_samples", ctypes.c_uint32),
        ("temporal_sample_spacing", ctypes.c_uint32),
        ("digitizer_gain", ctypes.c_double),
        ("digitizer_offset", ctypes.c_double),
    ]

    @staticmethod
    def size():
        return ctypes.sizeof(WaveformPacketStruct)


class WaveformPacketVlr(BaseKnownVLR):
    def __init__(self, record_id, description=""):
        super().__init__(record_id=record_id, description=description)
        self.parsed_record = None

    def parse_record_data(self, record_data):
        self.parsed_record = WaveformPacketStruct.from_buffer_copy(record_data)

    def record_data_bytes(self):
        return bytes(self.parsed_record)

    @staticmethod
    def official_record_ids():
        return range(100, 356)

    @staticmethod
    def official_user_id():
        return "LASF_Spec"

    @classmethod
    def from_raw(cls, raw_vlr):
        vlr = cls(
            raw_vlr.header.record_id, description=raw_vlr.header.description.decode()
        )
        vlr.description = raw_vlr.header.description
        vlr.parse_record_data(raw_vlr.record_data)
        return vlr


class GeoKeyEntryStruct(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("id", ctypes.c_uint16),
        ("tiff_tag_location", ctypes.c_uint16),
        ("count", ctypes.c_uint16),
        ("value_offset", ctypes.c_uint16),
    ]

    @staticmethod
    def size():
        return ctypes.sizeof(GeoKeysHeaderStructs)

    def __repr__(self):
        return "<GeoKey(Id: {}, Location: {}, count: {}, offset: {})>".format(
            self.id, self.tiff_tag_location, self.count, self.value_offset
        )


class GeoKeysHeaderStructs(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("key_direction_version", ctypes.c_uint16),
        ("key_revision", ctypes.c_uint16),
        ("minor_revision", ctypes.c_uint16),
        ("number_of_keys", ctypes.c_uint16),
    ]

    def __init__(self):
        super().__init__(
            key_directory_version=1, key_revision=1, minor_revision=0, number_of_kets=0
        )

    @staticmethod
    def size():
        return ctypes.sizeof(GeoKeysHeaderStructs)

    def __repr__(self):
        return "<GeoKeysHeader(vers: {}, rev:{}, minor: {}, num_keys: {})>".format(
            self.key_direction_version,
            self.key_revision,
            self.minor_revision,
            self.number_of_keys,
        )


class GeoKeyDirectoryVlr(BaseKnownVLR):
    def __init__(self):
        super().__init__(description="GeoTIFF GeoKeyDirectoryTag")
        self.geo_keys_header = GeoKeysHeaderStructs()
        self.geo_keys = [GeoKeyEntryStruct()]

    def parse_record_data(self, record_data):
        record_data = bytearray(record_data)
        header_data = record_data[: ctypes.sizeof(GeoKeysHeaderStructs)]
        self.geo_keys_header = GeoKeysHeaderStructs.from_buffer(header_data)
        self.geo_keys = []
        keys_data = record_data[GeoKeysHeaderStructs.size() :]
        num_keys = (
            len(record_data[GeoKeysHeaderStructs.size() :]) // GeoKeyEntryStruct.size()
        )
        if num_keys != self.geo_keys_header.number_of_keys:
            # print("Mismatch num keys")
            self.geo_keys_header.number_of_keys = num_keys

        for i in range(self.geo_keys_header.number_of_keys):
            data = keys_data[
                (i * GeoKeyEntryStruct.size()) : (i + 1) * GeoKeyEntryStruct.size()
            ]
            self.geo_keys.append(GeoKeyEntryStruct.from_buffer(data))

    def record_data_bytes(self):
        b = bytes(self.geo_keys_header)
        b += b"".join(map(bytes, self.geo_keys))
        return b

    def __repr__(self):
        return "<{}({} geo_keys)>".format(self.__class__.__name__, len(self.geo_keys))

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (34735,)


class GeoDoubleParamsVlr(BaseKnownVLR):
    def __init__(self):
        super().__init__(description="GeoTIFF GeoDoubleParamsTag")
        self.doubles = []

    def parse_record_data(self, record_data):
        sizeof_double = ctypes.sizeof(ctypes.c_double)
        if len(record_data) % sizeof_double != 0:
            raise ValueError(
                "GeoDoubleParams record data length () is not a multiple of sizeof(double) ()".format(
                    len(record_data), sizeof_double
                )
            )
        record_data = bytearray(record_data)
        num_doubles = len(record_data) // sizeof_double
        for i in range(num_doubles):
            b = record_data[i * sizeof_double : (i + 1) * sizeof_double]
            self.doubles.append(ctypes.c_double.from_buffer(b))

    def record_data_bytes(self):
        return b"".join(map(bytes, self.doubles))

    def __repr__(self):
        return "<GeoDoubleParamsVlr({})>".format(self.doubles)

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (34736,)


class GeoAsciiParamsVlr(BaseKnownVLR):
    def __init__(self):
        super().__init__(description="GeoTIFF GeoAsciiParamsTag")
        self.strings = []

    def parse_record_data(self, record_data):
        self.strings = [s.decode("ascii") for s in record_data.split(NULL_BYTE)]

    def record_data_bytes(self):
        return NULL_BYTE.join(s.encode("ascii") for s in self.strings)

    def __repr__(self):
        return "<GeoAsciiParamsVlr({})>".format(self.strings)

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (34737,)


class WktMathTransformVlr(BaseKnownVLR):
    """
    From the Spec:
        Note that the math transform WKT record is added for completeness, and a coordinate system WKT
        may or may not require a math transform WKT record

    """

    def __init__(self):
        super().__init__(description="")
        self.string = ""

    def _encode_string(self):
        return self.string.encode("utf-8") + NULL_BYTE

    def parse_record_data(self, record_data):
        self.string = record_data.decode("utf-8")

    def record_data_bytes(self):
        return self._encode_string()

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (2112,)


class WktCoordinateSystemVlr(BaseKnownVLR):
    """Replaces Coordinates Reference System for new las files (point fmt >= 5)
    "LAS is not using the “ESRI WKT”
    """

    def __init__(self, wkt_string=""):
        super().__init__(description="OGC Transformation Record")
        self.string = wkt_string

    def _encode_string(self):
        return self.string.encode("utf-8") + NULL_BYTE

    def parse_record_data(self, record_data):
        self.string = record_data.decode("utf-8")

    def record_data_bytes(self):
        return self._encode_string()

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (2112,)


def vlr_factory(raw_vlr):
    """Given a raw_vlr tries to find its corresponding KnownVLR class
    that can parse its data.
    If no KnownVLR implementation is found, returns a VLR (record_data will still be bytes)
    """
    user_id = raw_vlr.header.user_id.rstrip(NULL_BYTE).decode()
    known_vlrs = BaseKnownVLR.__subclasses__()
    for known_vlr in known_vlrs:
        if (
            known_vlr.official_user_id() == user_id
            and raw_vlr.header.record_id in known_vlr.official_record_ids()
        ):
            try:
                return known_vlr.from_raw(raw_vlr)
            except Exception as err:
                logger.warning(f"Failed to parse {known_vlr}: {err}")
                return VLR.from_raw(raw_vlr)

    return VLR.from_raw(raw_vlr)



""" Contains the classes that manages Las PointRecords
Las PointRecords are represented using Numpy's structured arrays,
The PointRecord classes provide a few extra things to manage these arrays
in the context of Las point data
"""
import logging
from abc import ABC, abstractmethod
from typing import NoReturn, Any, List, Tuple

import numpy as np

from . import dims, packing
from .dims import SubFieldView, ScaledArrayView
from .. import errors
from ..point import PointFormat

logger = logging.getLogger(__name__)


def scale_dimension(array_dim, scale, offset):
    return (array_dim * scale) + offset


def unscale_dimension(array_dim, scale, offset):
    return np.round((np.array(array_dim) - offset) / scale)


def raise_not_enough_bytes_error(
        expected_bytes_len, missing_bytes_len, point_data_buffer_len, points_dtype
) -> NoReturn:
    raise errors.PylasError(
        "The file does not contain enough bytes to store the expected number of points\n"
        "expected {} bytes, read {} bytes ({} bytes missing == {} points) and it cannot be corrected\n"
        "{} (bytes) / {} (point_size) = {} (points)".format(
            expected_bytes_len,
            point_data_buffer_len,
            missing_bytes_len,
            missing_bytes_len / points_dtype.itemsize,
            point_data_buffer_len,
            points_dtype.itemsize,
            point_data_buffer_len / points_dtype.itemsize,
        )
    )


class IPointRecord(ABC):
    """Wraps the numpy structured array contained the points data"""

    @property
    @abstractmethod
    def point_format(self) -> PointFormat:
        ...

    @property
    @abstractmethod
    def array(self) -> np.ndarray:
        ...

    @property
    @abstractmethod
    def point_size(self):
        """Shall return the point size as that will be written in the header"""
        return self.point_format.size

    @property
    @abstractmethod
    def actual_point_size(self):
        """Shall return the actual size in bytes that the points take in memory"""
        return self.array.dtype.itemsize

    @abstractmethod
    def __getitem__(self, item):
        pass

    @abstractmethod
    def __setitem__(self, key, value):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def raw_bytes(self):
        pass

    @abstractmethod
    def write_to(self, out):
        pass

    @classmethod
    @abstractmethod
    def from_stream(cls, stream, point_format_id, count):
        pass

    @classmethod
    @abstractmethod
    def empty(cls, point_format_id):
        pass


class PointRecord(IPointRecord, ABC):
    def __init__(self, data, point_format: PointFormat):
        self._array = data
        self._point_format = point_format

    @property
    def point_format(self) -> PointFormat:
        return self._point_format

    @property
    def array(self) -> np.ndarray:
        return self._array

    @property
    def dimensions_names(self):
        return self.point_format.dimension_names

    @property
    def extra_dimensions_names(self):
        """Returns the names of extra-dimensions contained in the PointRecord"""
        return self.point_format.extra_dimension_names

    @property
    def actual_point_size(self):
        """Returns the point size in bytes taken by each points of the record

        Returns
        -------
        int
            The point size in byte

        """
        return self.array.dtype.itemsize

    @classmethod
    def from_point_record(cls, other_point_record, new_point_format):
        """Construct a new PackedPointRecord from an existing one with the ability to change
        to point format while doing so
        """
        array = np.zeros_like(other_point_record.array, dtype=new_point_format.dtype())
        new_record = cls(array, new_point_format)
        new_record.copy_fields_from(other_point_record)
        return new_record

    def copy_fields_from(self, other_record: 'PointRecord'):
        """Tries to copy the values of the current dimensions from other_record"""
        for dim_name in self.dimensions_names:
            try:
                self[dim_name] = np.array(other_record[dim_name])
            except ValueError:
                pass

    def copy_fields_from_numpy_array(self, array: np.ndarray):
        for dim_name in self.array.dtype.names:
            try:
                self[dim_name] = array[dim_name]
            except ValueError:
                pass

    def add_extra_dims(self, dim_tuples: List[Tuple[str, str]]):
        for name, type_str in dim_tuples:
            self.point_format.add_extra_dimension(name, type_str)
        old_array = self.array
        self._array = np.zeros_like(old_array, dtype=self.point_format.dtype())
        self.copy_fields_from_numpy_array(old_array)

    def add_extra_dim(self, name, type_str):
        self.add_extra_dims([(name, type_str)])

    def memoryview(self):
        return memoryview(self.array)

    def raw_bytes(self):
        return self.array.tobytes()

    def __getitem__(self, item):
        return self.array[item]

    def __setitem__(self, key, value):
        self._append_zeros_if_too_small(value)
        self.array[key] = np.array(value)

    def resize(self, new_size):
        size_diff = new_size - len(self.array)
        if size_diff > 0:
            self._array = np.append(
                self.array, np.zeros(size_diff, dtype=self.array.dtype)
            )
        elif size_diff < 0:
            self._array = self._array[:new_size].copy()

    def _append_zeros_if_too_small(self, value):
        """Appends zeros to the points stored if the value we are trying to
        fit is bigger
        """
        size_diff = len(value) - len(self.array)
        if size_diff > 0:
            self.resize(size_diff)

    def __eq__(self, other):
        return self.point_format == other.point_format and np.all(
            self.array == other.array
        )

    def __getattr__(self, item):
        try:
            return self[item]
        except ValueError:
            raise AttributeError("{} is not a valid dimension".format(item)) from None

    def __len__(self):
        return self.array.shape[0]


class PackedPointRecord(PointRecord):
    """
    In the PackedPointRecord, fields that are a combinations of many sub-fields (fields stored on less than a byte)
    are still packed together and are only de-packed and re-packed when accessed.

    This uses of less memory than if the sub-fields were unpacked

    >>> #return number is a sub-field
    >>> from pylas import PointFormat
    >>> packed_point_record = PackedPointRecord.zeros(PointFormat(0), 10)
    >>> return_number = packed_point_record['return_number']
    >>> return_number
    <SubFieldView([0 0 0 0 0 0 0 0 0 0])>
    >>> return_number[:] = 1
    >>> np.alltrue(packed_point_record['return_number'] == 1)
    True
    """

    def __init__(self, data, point_format):
        super().__init__(data, point_format)
        self.sub_fields_dict = dims.get_sub_fields_dict(point_format.id)

    @property
    def all_dimensions_names(self):
        """Returns all the dimensions names, including the names of sub_fields
        and their corresponding packed fields
        """
        return frozenset(self.array.dtype.names + tuple(self.sub_fields_dict.keys()))

    @property
    def point_size(self):
        """Returns the point size in bytes taken by each points of the record

        Returns
        -------
        int
            The point size in byte

        """
        return self.array.dtype.itemsize

    @classmethod
    def zeros(cls, point_format, point_count):
        """Creates a new point record with all dimensions initialized to zero

        Parameters
        ----------
        point_format: PointFormat
            The point format id the point record should have
        point_count : int
            The number of point the point record should have

        Returns
        -------
        PackedPointRecord

        """
        data = np.zeros(point_count, point_format.dtype())
        return cls(data, point_format)

    @classmethod
    def empty(cls, point_format):
        """Creates an empty point record.

        Parameters
        ----------
        point_format: pylas.PointFormat
            The point format id the point record should have

        Returns
        -------
        PackedPointRecord

        """
        return cls.zeros(point_format, point_count=0)

    @classmethod
    def from_stream(cls, stream, point_format, count):
        """Construct the point record by reading the points from the stream"""
        points_dtype = point_format.dtype
        point_data_buffer = bytearray(stream.read(count * points_dtype.itemsize))

        try:
            data = np.frombuffer(point_data_buffer, dtype=points_dtype, count=count)
        except ValueError:
            expected_bytes_len = count * points_dtype.itemsize
            if len(point_data_buffer) % points_dtype.itemsize != 0:
                missing_bytes_len = expected_bytes_len - len(point_data_buffer)
                raise_not_enough_bytes_error(
                    expected_bytes_len,
                    missing_bytes_len,
                    len(point_data_buffer),
                    points_dtype,
                )
            else:
                actual_count = len(point_data_buffer) // points_dtype.itemsize
                logger.critical(
                    "Expected {} points, there are {} ({} missing)".format(
                        count, actual_count, count - actual_count
                    )
                )
                data = np.frombuffer(
                    point_data_buffer, dtype=points_dtype, count=actual_count
                )

        return cls(data, point_format)

    @classmethod
    def from_buffer(cls, buffer, point_format, count, offset=0):
        points_dtype = point_format.dtype()
        data = np.frombuffer(buffer, dtype=points_dtype, offset=offset, count=count)

        return cls(data, point_format)

    def write_to(self, out):
        """ Writes the points to the output stream"""
        out.write(self.raw_bytes())

    def __getitem__(self, item):
        """Gives access to the underlying numpy array
        Unpack the dimension if item is the name a sub-field
        """
        try:
            composed_dim, sub_field = self.sub_fields_dict[item]
            return dims.SubFieldView(self.array[composed_dim], sub_field.mask)
        except KeyError:
            return self.array[item]

    def __setitem__(self, key, value):
        """Sets elements in the array"""
        self._append_zeros_if_too_small(value)
        try:
            composed_dim, sub_field = self.sub_fields_dict[key]
            if isinstance(value, SubFieldView):
                value = np.array(value)
            try:
                packing.pack(
                    self.array[composed_dim], value, sub_field.mask, inplace=True
                )
            except OverflowError as e:
                raise OverflowError(
                    "Overflow when packing {} into {}: {}".format(
                        sub_field.name, composed_dim, e
                    )
                )
        except KeyError:
            self.array[key] = np.array(value)

    def __repr__(self):
        return "<PackedPointRecord(fmt: {}, len: {}, point size: {})>".format(
            self.point_format, len(self), self.actual_point_size
        )


def apply_new_scaling(record, scales, offsets) -> None:
    record['X'] = unscale_dimension(np.asarray(record.x), scales[0], offsets[0])
    record['Y'] = unscale_dimension(np.asarray(record.y), scales[1], offsets[1])
    record['Z'] = unscale_dimension(np.asarray(record.x), scales[2], offsets[2])


class ScaleAwarePointRecord(PackedPointRecord):
    def __init__(self, array, point_format, scales, offsets):
        super().__init__(array, point_format)
        self.scales = scales
        self.offsets = offsets

    def change_scaling(self, scales=None, offsets=None) -> None:
        if scales is not None:
            self.scales = scales
        if offsets is not None:
            self.offsets = offsets

        apply_new_scaling(self, scales, offsets)

        self.scales = scales
        self.offsets = offsets

    def __getitem__(self, item):
        if isinstance(item, (slice, np.ndarray)):
            return ScaleAwarePointRecord(
                self.array[item], self.point_format, self.scales, self.offsets
            )

        if item == "x":
            return ScaledArrayView(self.array["X"], self.scales[0], self.offsets[0])
        elif item == "y":
            return ScaledArrayView(self.array["Y"], self.scales[1], self.offsets[1])
        elif item == "z":
            return ScaledArrayView(self.array["Z"], self.scales[2], self.offsets[2])
        else:
            return super().__getitem__(item)




""" 'Entry point' of the library, Contains the various functions meant to be
used directly by a user
"""
import copy
import io
import logging
import os
from pathlib import Path
from typing import Union

import numpy as np

from . import headers, utils
from .compression import LazBackend
from .errors import PylasError
from .lasdatas import las12, las14
from .lasmmap import LasMMAP
from .lasreader import LasReader
from .laswriter import LasWriter
from .lasappender import LasAppender
from .point import dims, record, PointFormat

logger = logging.getLogger(__name__)


def open_las(
    source,
    mode="r",
    closefd=True,
    laz_backend=LazBackend.detect_available(),
    header=None,
    do_compress=None,
) -> Union[LasReader, LasWriter, LasAppender]:
    """Opens and reads the header of the las content in the source

        >>> with open_las('pylastests/simple.las') as f:
        ...     print(f.header.point_format_id)
        3


        >>> f = open('pylastests/simple.las', mode='rb')
        >>> with open_las(f,closefd=False) as flas:
        ...     print(flas.header)
        <LasHeader(1.2)>
        >>> f.closed
        False
        >>> f.close()
        >>> f.closed
        True


        >>> f = open('pylastests/simple.las', mode='rb')
        >>> with open_las(f) as flas:
        ...    las = flas.read()
        >>> f.closed
        True

    Parameters
    ----------
    source: str or bytes or io.BytesIO
        if source is a str it must be a filename
        a stream if a file object with the methods read, seek, tell

    mode: Optional, the mode to open the file "r" for reading, "w" for writing
          "r" by default

    laz_backend: Optional, the LAZ backend to use to handle decompression/comression
                 By default available_backends are detected, see LazBacked to see the
                 preference order when multiple backends are available

    header: The header to use when opening in write mode.

    do_compress: optional, bool
                 -> None (default) guess if compression is needed using the file extension
                 -> True compresses the file
                 -> False do not compress the file

    closefd: bool
        Whether the stream/file object shall be closed, this only work
        when using open_las in a with statement. An exception is raised if
        closefd is specified and the source is a filename


    Returns
    -------
    pylas.lasreader.LasReader

    """
    if mode == "r":
        if header is not None:
            raise PylasError(
                "header argument is not used when opening in read mode, "
                "did you meant to open in write mode ?"
            )
        if do_compress is not None:
            raise PylasError(
                "do_compress argument is not used when opening in read mode, "
                "did you meant to open in write mode ?"
            )
        if isinstance(source, (str, Path)):
            stream = open(source, mode="rb", closefd=closefd)
        elif isinstance(source, bytes):
            stream = io.BytesIO(source)
        else:
            stream = source
        return LasReader(stream, closefd=closefd, laz_backend=laz_backend)
    elif mode == "w":
        if header is None:
            raise ValueError("A header is needed when opening a file for writing")

        if isinstance(source, (str, Path)):
            if do_compress is None:
                do_compress = os.path.splitext(source)[1].lower() == ".laz"
            stream = open(source, mode="wb+", closefd=closefd)
        elif isinstance(source, bytes):
            stream = io.BytesIO(source)
        else:
            assert source.seekable()
            stream = source
        if do_compress is None:
            do_compress = False
        return LasWriter(
            stream,
            header=header,
            do_compress=do_compress,
            laz_backend=laz_backend,
            closefd=closefd,
        )
    elif mode == "a":
        if isinstance(source, (str, Path)):
            stream = open(source, mode="rb+", closefd=closefd)
        elif isinstance(source, bytes):
            stream = io.BytesIO(source)
        else:
            stream = source
        return LasAppender(stream, closefd=closefd, laz_backend=laz_backend)

    else:
        raise ValueError("Unknown mode '{}'".format(mode))


def read_las(source, closefd=True, laz_backend=LazBackend.detect_available()):
    """Entry point for reading las data in pylas

    Reads the whole file into memory.

    >>> las = read_las("pylastests/simple.las")
    >>> las.classification
    <SubFieldView([1 1 1 ... 1 1 1])>

    Parameters
    ----------
    source : str or io.BytesIO
        The source to read data from

    laz_backend: Optional, the backend to use when the file is as LAZ file.
                 By default pylas will find the backend to use by himself.
                 Use if you wan a specific backend to be used

    closefd: bool
            if True and the source is a stream, the function will close it
            after it is done reading


    Returns
    -------
    pylas.lasdatas.base.LasBase
        The object you can interact with to get access to the LAS points & VLRs
    """
    with open_las(source, closefd=closefd, laz_backend=laz_backend) as reader:
        return reader.read()


def mmap_las(filename):
    """MMap a file, much like laspy did, very experimental
    not well tested
    """
    return LasMMAP(filename)


def create_from_header(header):
    """Creates a File from an existing header,
    allocating the array of point according to the provided header.
    The input header is copied.


    Parameters
    ----------
    header : existing header to be used to create the file

    Returns
    -------
    pylas.lasdatas.base.LasBase
    """
    header = copy.copy(header)
    header.point_count = 0
    points = record.PackedPointRecord.empty(PointFormat(header.point_format_id))
    if header.version >= "1.4":
        return las14.LasData(header=header, points=points)
    return las12.LasData(header=header, points=points)


def create_las(*, point_format_id=0, file_version=None):
    """Function to create a new empty las data object

    .. note::

        If you provide both point_format and file_version
        an exception will be raised if they are not compatible

    >>> las = create_las(point_format_id=6,file_version="1.2")
    Traceback (most recent call last):
     ...
    pylas.errors.PylasError: Point format 6 is not compatible with file version 1.2


    If you provide only the point_format the file_version will automatically
    selected for you.

    >>> las = create_las(point_format_id=0)
    >>> las.header.version == '1.2'
    True

    >>> las = create_las(point_format_id=6)
    >>> las.header.version == '1.4'
    True


    Parameters
    ----------
    point_format_id: int
        The point format you want the created file to have

    file_version: str, optional, default=None
        The las version you want the created las to have

    Returns
    -------
    pylas.lasdatas.base.LasBase
       A new las data object

    """
    if file_version is not None:
        dims.raise_if_version_not_compatible_with_fmt(point_format_id, file_version)
    else:
        file_version = dims.min_file_version_for_point_format(point_format_id)

    header = headers.HeaderFactory.new(file_version)
    header.point_format_id = point_format_id

    if file_version >= "1.4":
        return las14.LasData(header=header)
    return las12.LasData(header=header)


def convert(source_las, *, point_format_id=None, file_version=None):
    """Converts a Las from one point format to another
    Automatically upgrades the file version if source file version is not compatible with
    the new point_format_id


    convert to point format 0

    >>> las = read_las('pylastests/simple.las')
    >>> las.header.version
    '1.2'
    >>> las = convert(las, point_format_id=0)
    >>> las.header.point_format_id
    0
    >>> las.header.version
    '1.2'

    convert to point format 6, which need version >= 1.4
    then convert back to point format 0, version is not downgraded

    >>> las = read_las('pylastests/simple.las')
    >>> las.header.version
    '1.2'
    >>> las = convert(las, point_format_id=6)
    >>> las.header.point_format_id
    6
    >>> las.header.version
    '1.4'
    >>> las = convert(las, point_format_id=0)
    >>> las.header.version
    '1.4'

    an exception is raised if the requested point format is not compatible
    with the file version

    >>> las = read_las('pylastests/simple.las')
    >>> convert(las, point_format_id=6, file_version='1.2')
    Traceback (most recent call last):
     ...
    pylas.errors.PylasError: Point format 6 is not compatible with file version 1.2

    Parameters
    ----------
    source_las : pylas.lasdatas.base.LasBase
        The source data to be converted

    point_format_id : int, optional
        The new point format id (the default is None, which won't change the source format id)

    file_version : str, optional,
        The new file version. None by default which means that the file_version
        may be upgraded for compatibility with the new point_format. The file version will not
        be downgraded.

    Returns
    -------
        pylas.lasdatas.base.LasBase
    """
    if point_format_id is None:
        point_format_id = source_las.point_format.id

    if file_version is None:
        file_version = max(
            source_las.header.version,
            dims.min_file_version_for_point_format(point_format_id),
        )
    else:
        file_version = str(file_version)
        dims.raise_if_version_not_compatible_with_fmt(point_format_id, file_version)

    header = headers.HeaderFactory.convert_header(source_las.header, file_version)
    header.point_format_id = point_format_id

    point_format = PointFormat(point_format_id)
    point_format.dimensions.extend(source_las.point_format.extra_dimensions)
    points = record.PackedPointRecord.from_point_record(source_las.points, point_format)

    try:
        evlrs = source_las.evlrs
    except AttributeError:
        evlrs = []

    if file_version >= "1.4":

        las = las14.LasData(
            header=header, vlrs=source_las.vlrs, points=points, evlrs=evlrs
        )
    else:
        if evlrs:
            logger.warning(
                "The source contained {} EVLRs,"
                " they will be lost as version {} doest not support them".format(
                    len(evlrs), file_version
                )
            )
        las = las12.LasData(header=header, vlrs=source_las.vlrs, points=points)

    return las


def merge_las(*las_files):
    """Merges multiple las files into one

    merged = merge_las(las_1, las_2)
    merged = merge_las([las_1, las_2, las_3])

    Parameters
    ----------
    las_files: Iterable of LasData or LasData

    Returns
    -------
    pylas.lasdatas.base.LasBase
        The result of the merging

    """
    if len(las_files) == 1:
        las_files = las_files[0]

    if not las_files:
        raise ValueError("No files to merge")

    if not utils.files_have_same_dtype(las_files):
        raise ValueError("All files must have the same point format")

    header = las_files[0].header
    num_pts_merged = sum(len(las.points) for las in las_files)

    # scaled x, y, z have to be set manually
    # to be sure to have a good offset in the header
    merged = create_from_header(header)
    # TODO extra dimensions should be managed better here

    for dim_name, dim_type in las_files[0].points.point_format.extra_dims:
        merged.add_extra_dim(dim_name, dim_type)

    merged.points = np.zeros(num_pts_merged, merged.points.dtype)
    merged_x = np.zeros(num_pts_merged, np.float64)
    merged_y = np.zeros(num_pts_merged, np.float64)
    merged_z = np.zeros(num_pts_merged, np.float64)

    offset = 0
    for i, las in enumerate(las_files, start=1):
        slc = slice(offset, offset + len(las.points))
        merged.points[slc] = las.points
        merged_x[slc] = las.x
        merged_y[slc] = las.y
        merged_z[slc] = las.z
        merged["point_source_id"][slc] = i
        offset += len(las.points)

    merged.x = merged_x
    merged.y = merged_y
    merged.z = merged_z

    return merged


def write_then_read_again(
    las, do_compress=False, laz_backend=LazBackend.detect_available()
):
    """writes the given las into memory using BytesIO and
    reads it again, returning the newly read file.

    Mostly used for testing purposes, without having to write to disk
    """
    out = io.BytesIO()
    las.write(out, do_compress=do_compress, laz_backend=laz_backend)
    out.seek(0)
    return read_las(out)



import io
import math
from typing import Union, Iterable, BinaryIO

from .headers.rawheader import Header
from .vlrs.vlrlist import VLRList
from .compression import LazBackend
from .errors import PylasError
from .evlrs import EVLRList, RawEVLRList
from .lasreader import LasReader, get_extra_dims_info_tuple
from .laswriter import UncompressedPointWriter
from .point.format import PointFormat
from .point.record import PointRecord

try:
    import lazrs
except ModuleNotFoundError:
    pass


class LazrsAppender:
    """Appending in LAZ file works by seeking to start of the last chunk
    of compressed points, decompress it while keeping the points in
    memory.

    Then seek back to the start of the last chunk, and recompress
    the points we just read, so that we have a compressor in the proper state
    ready to compress new points.
    """

    def __init__(
        self, dest: BinaryIO, header: Header, vlrs: VLRList, parallel: bool
    ) -> None:
        self.dest = dest
        self.offset_to_point_data = header.offset_to_point_data
        laszip_vlr = vlrs.pop(vlrs.index("LasZipVlr"))

        self.dest.seek(header.offset_to_point_data, io.SEEK_SET)
        decompressor = lazrs.LasZipDecompressor(self.dest, laszip_vlr.record_data)
        vlr = decompressor.vlr()
        number_of_complete_chunk = int(
            math.floor(header.point_count / vlr.chunk_size())
        )

        self.dest.seek(header.offset_to_point_data, io.SEEK_SET)
        chunk_table = lazrs.read_chunk_table(self.dest)
        if chunk_table is None:
            # The file does not have a chunk table
            # we cannot seek to the last chunk, so instead, we will
            # decompress points (which is slower) and build the chunk table
            # to write it later

            self.chunk_table = []
            start_of_chunk = self.dest.tell()
            point_buf = bytearray(vlr.chunk_size() * vlr.item_size())

            for _ in range(number_of_complete_chunk):
                decompressor.decompress_many(point_buf)
                pos = self.dest.tell()
                self.chunk_table.append(pos - start_of_chunk)
                start_of_chunk = pos
        else:
            self.chunk_table = chunk_table[:-1]
            idx_first_point_of_last_chunk = number_of_complete_chunk * vlr.chunk_size()
            decompressor.seek(idx_first_point_of_last_chunk)

        points_of_last_chunk = bytearray(
            (header.point_count % vlr.chunk_size()) * vlr.item_size()
        )
        decompressor.decompress_many(points_of_last_chunk)

        self.dest.seek(header.offset_to_point_data, io.SEEK_SET)
        if parallel:
            self.compressor = lazrs.ParLasZipCompressor(
                self.dest, vlr
            )  # This overwrites old offset
        else:
            self.compressor = lazrs.LasZipCompressor(
                self.dest, vlr
            )  # This overwrites the old offset
        self.dest.seek(sum(self.chunk_table), io.SEEK_CUR)
        self.compressor.compress_many(points_of_last_chunk)

    def write_points(self, points: PointRecord) -> None:
        self.compressor.compress_many(points.memoryview())

    def done(self) -> None:
        # The chunk table written is at the good position
        # but it is incomplete (it's missing the chunk_table of
        # chunks before the one we appended)
        self.compressor.done()

        # So we update it
        self.dest.seek(self.offset_to_point_data, io.SEEK_SET)
        offset_to_chunk_table = int.from_bytes(self.dest.read(8), "little", signed=True)
        self.dest.seek(-8, io.SEEK_CUR)
        chunk_table = self.chunk_table + lazrs.read_chunk_table(self.dest)
        self.dest.seek(offset_to_chunk_table, io.SEEK_SET)
        lazrs.write_chunk_table(self.dest, chunk_table)


class LasAppender:
    def __init__(
        self,
        dest: BinaryIO,
        laz_backend: Union[LazBackend, Iterable[LazBackend]] = (
            LazBackend.LazrsParallel,
            LazBackend.Lazrs,
        ),
        closefd: bool = True,
    ) -> None:
        if not dest.seekable():
            raise TypeError("Expected the 'dest' to be a seekable file object")
        header, vlrs = LasReader._read_header_and_vlrs(dest, seekable=True)

        self.dest = dest
        self.header = header
        self.vlrs = vlrs
        self.point_format = PointFormat(
            self.header.point_format_id,
            get_extra_dims_info_tuple(self.header, self.vlrs),
        )

        if not header.are_points_compressed:
            self.points_writer = UncompressedPointWriter(self.dest)
            self.dest.seek(
                (self.header.point_count * self.header.point_size)
                + self.header.offset_to_point_data,
                io.SEEK_SET,
            )
        else:
            self.points_writer = self._create_laz_backend(laz_backend)

        if header.version >= "1.4" and header.number_of_evlr > 0:
            assert (
                self.dest.tell() <= self.header.start_of_first_evlr
            ), "The position is past the start of evlrs"
            pos = self.dest.tell()
            self.dest.seek(self.header.start_of_first_evlr, io.SEEK_SET)
            self.evlrs = EVLRList.read_from(self.dest, self.header.number_of_evlr)
            dest.seek(self.header.start_of_first_evlr, io.SEEK_SET)
            self.dest.seek(pos, io.SEEK_SET)
        elif header.version >= "1.4":
            self.evlrs = []

        self.closefd = closefd

    def append_points(self, points: PointRecord) -> None:
        if points.point_format != self.point_format:
            raise PylasError("Point formats do not match")

        self.points_writer.write_points(points)
        self.header.update(points)

    def close(self) -> None:
        self.points_writer.done()
        self._write_evlrs()
        self._write_updated_header()

        if self.closefd:
            self.dest.close()

    def _write_evlrs(self) -> None:
        if self.header.version >= "1.4" and len(self.evlrs) > 0:
            self.header.number_of_evlr = len(self.evlrs)
            self.header.start_of_first_evlr = self.dest.tell()
            raw_evlrs = RawEVLRList.from_list(self.evlrs)
            raw_evlrs.write_to(self.dest)

    def _write_updated_header(self) -> None:
        pos = self.dest.tell()
        self.dest.seek(0, io.SEEK_SET)
        self.header.write_to(self.dest)
        self.dest.seek(pos, io.SEEK_SET)

    def _create_laz_backend(
        self,
        laz_backend: Union[LazBackend, Iterable[LazBackend]] = (
            LazBackend.LazrsParallel,
            LazBackend.Lazrs,
        ),
    ) -> LazrsAppender:
        try:
            laz_backend = iter(laz_backend)
        except TypeError:
            laz_backend = (laz_backend,)

        last_error = None
        for backend in laz_backend:
            if backend == LazBackend.Laszip:
                raise PylasError("Laszip backend does not support appending")
            elif backend == LazBackend.LazrsParallel:
                try:
                    return LazrsAppender(
                        self.dest, self.header, self.vlrs, parallel=True
                    )
                except Exception as e:
                    last_error = e
            elif backend == LazBackend.Lazrs:
                try:
                    return LazrsAppender(
                        self.dest, self.header, self.vlrs, parallel=False
                    )
                except Exception as e:
                    last_error = e

        if last_error is not None:
            raise PylasError(f"Could not initialize a laz backend: {last_error}")
        else:
            raise PylasError(f"No valid laz backend selected")

    def __enter__(self) -> "LasAppender":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()



from itertools import zip_longest
from typing import Tuple, Optional, Iterable

import numpy as np

from . import dims
from .dims import DimensionInfo


class PointFormat:
    """Class that contains the dimensions that forms a PointFormat

    A PointFormat has 'standard' dimensions (dimensions defined in the LAS standard, each
    point format has its set of dimensions), but it can also have extra (non-standard) dimensions
    defined by the user)

    This class can be used to get information about dimensions of a point format.


    >>> fmt = PointFormat(3)
    >>> all(dim.is_standard for dim in fmt.dimensions)
    True
    >>> dim = fmt.dimension_by_name("classification") # or fmt["classification"]
    >>> dim.max
    31
    >>> dim.min
    0
    >>> dim.num_bits
    5

    """

    def __init__(
        self,
        point_format_id: int,
        extra_dims: Optional[Tuple[Tuple[str, str], ...]] = None,
    ):
        """
        Parameters
        ----------
        point_format_id: int
            point format id
        extra_dims: list of tuple
            [(name, type_str), ..] of extra dimensions attached to this point format
        """
        self.id = point_format_id
        self.dimensions = []
        composed_dims = dims.COMPOSED_FIELDS[self.id]
        for dim_name in dims.ALL_POINT_FORMATS_DIMENSIONS[self.id]:
            try:
                sub_fields = composed_dims[dim_name]
            except KeyError:
                dimension = DimensionInfo.from_type_str(
                    dim_name, dims.DIMENSIONS_TO_TYPE[dim_name], is_standard=True
                )
                self.dimensions.append(dimension)
            else:
                for sub_field in sub_fields:
                    dimension = DimensionInfo.from_bitmask(
                        sub_field.name, sub_field.mask, is_standard=True
                    )
                    self.dimensions.append(dimension)

        if extra_dims is not None:
            for name, type_str in extra_dims:
                self.add_extra_dimension(name, type_str)

    @property
    def standard_dimensions(self) -> Iterable[DimensionInfo]:
        return (dim for dim in self.dimensions if dim.is_standard)

    @property
    def extra_dimensions(self) -> Iterable[DimensionInfo]:
        return (dim for dim in self.dimensions if dim.is_standard is False)

    @property
    def dimension_names(self) -> Iterable[str]:
        """Returns the names of the dimensions contained in the point format"""
        return (dim.name for dim in self.dimensions)

    @property
    def standard_dimension_names(self) -> Iterable[str]:
        """Returns the names of the extra dimensions in this point format"""
        return (dim.name for dim in self.standard_dimensions)

    @property
    def extra_dimension_names(self) -> Iterable[str]:
        """Returns the names of the extra dimensions in this point format"""
        return (dim.name for dim in self.extra_dimensions)

    @property
    def size(self) -> int:
        """Returns the number of bytes (standard + extra)"""
        return int(sum(dim.num_bits for dim in self.dimensions) // 8)

    @property
    def num_standard_bytes(self) -> int:
        """Returns the number of bytes used by standard dims"""
        return int(sum(dim.num_bits for dim in self.standard_dimensions) // 8)

    @property
    def num_extra_bytes(self) -> int:
        """Returns the number of extra bytes"""
        return int(sum(dim.num_bits for dim in self.extra_dimensions) // 8)

    @property
    def has_waveform_packet(self):
        """Returns True if the point format has waveform packet dimensions"""
        dimensions = set(self.dimension_names)
        return all(name in dimensions for name in dims.WAVEFORM_FIELDS_NAMES)

    def dimension_by_name(self, name: str) -> DimensionInfo:
        for dim in self.dimensions:
            if dim.name == name:
                return dim
        raise ValueError(f"Dimension '{name}' does not exist")

    def add_extra_dimension(self, name: str, type_str: str) -> None:
        self.dimensions.append(
            DimensionInfo.from_type_str(name, type_str, is_standard=False)
        )

    def dtype(self):
        """Returns the numpy.dtype used to store the point records in a numpy array

        .. note::

            The dtype corresponds to the dtype with sub_fields *packed* into their
            composed fields

        """
        dtype = dims.ALL_POINT_FORMATS_DTYPE[self.id]
        descr = dtype.descr
        for extra_dim in self.extra_dimensions:
            descr.append((extra_dim.name, extra_dim.type_str()))
        return np.dtype(descr)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.dimension_by_name(item)
        return self.dimensions[item]

    def __eq__(self, other):
        if self.id != other.id:
            return False

        for my_eb, ot_eb in zip_longest(self.extra_dimensions, other.extra_dimensions):
            if my_eb != ot_eb:
                return False

        return True

    def __repr__(self):
        return "<PointFormat({}, {} bytes of extra dims)>".format(
            self.id, self.num_extra_bytes
        )


def lost_dimensions(point_fmt_in, point_fmt_out):
    """Returns a list of the names of the dimensions that will be lost
    when converting from point_fmt_in to point_fmt_out
    """

    dimensions_in = set(PointFormat(point_fmt_in).dimension_names)
    dimensions_out = set(PointFormat(point_fmt_out).dimension_names)

    completely_lost = []
    for dim_name in dimensions_in:
        if dim_name not in dimensions_out:
            completely_lost.append(dim_name)
    return completely_lost



""" All the custom exceptions types
"""


class PylasError(Exception):
    pass


class UnknownExtraType(PylasError):
    pass


class PointFormatNotSupported(PylasError):
    pass


class FileVersionNotSupported(PylasError):
    pass


class LazError(PylasError):
    pass


class IncompatibleDataFormat(PylasError):
    pass



import pathlib
from typing import Union

from .compression import LazBackend
from .lasdatas import las12, las14

LazBackend = LazBackend

LasData = Union[las12.LasData, las14.LasData]

PathLike = Union[str, pathlib.Path]



import threading


def ctypes_max_limit(byte_size, signed=False):
    nb_bits = (byte_size * 8) - (1 if signed else 0)
    return (2 ** nb_bits) - 1


def files_have_same_point_format_id(las_files):
    """Returns true if all the files have the same points format id"""
    point_format_found = {las.header.point_format_id for las in las_files}
    return len(point_format_found) == 1


def files_have_same_dtype(las_files):
    """Returns true if all the files have the same numpy datatype"""
    dtypes = {las.points.dtype for las in las_files}
    return len(dtypes) == 1


class ConveyorThread(threading.Thread):
    """class to be used as a separate thread by calling start()

    This class convey data from the input stream into the output stream.
    This is used when piping data into laszip.exe using python's subprocess.Popen
    when both of the stdin & stdout are in memory io objects because in such cases
    there is a deadlock occuring because we fill up the os stdout/stdin buffer

    So we need a separate thread to convey / move data from one source into another
    so that the main thread can read and we avoid deadlocks
    """

    def __init__(self, input_stream, output_stream, close_output=False):
        super().__init__()
        self.input_stream = input_stream
        self.output_stream = output_stream
        self.close_output = close_output
        self.should_end = False

    def run(self) -> None:
        for data in self.input_stream:
            if data:
                self.output_stream.write(data)
            elif self.should_end:
                break
            else:
                break

        if self.close_output:
            self.output_stream.close()

    def ask_for_termination(self) -> None:
        self.should_end = True



""" This module contains functions to pack and unpack point dimensions
"""
import numpy as np


def least_significant_bit_set(mask: int) -> int:
    """Return the least significant bit set

    The index is 0-indexed.
    Returns -1 is no bit is set

    >>> least_significant_bit_set(0b0000_0001)
    0
    >>> least_significant_bit_set(0b0001_0000)
    4
    >>> least_significant_bit_set(0b0000_0000)
    -1
    """
    return (mask & -mask).bit_length() - 1


def pack(array, sub_field_array, mask, inplace=False):
    """Packs a sub field's array into another array using a mask

    Parameters:
    ----------
    array : numpy.ndarray
        The array in which the sub field array will be packed into
    array_in : numpy.ndarray
        sub field array to pack
    mask : mask (ie: 0b00001111)
        Mask of the sub field
    inplace : {bool}, optional
        If true a new array is returned. (the default is False, which modifies the array in place)

    Raises
    ------
    OverflowError
        If the values contained in the sub field array are greater than its mask's number of bits
        allows
    """
    lsb = least_significant_bit_set(mask)
    max_value = int(mask >> lsb)
    if np.max(sub_field_array) > max_value:
        raise OverflowError(
            "value ({}) is greater than allowed (max: {})".format(
                sub_field_array.max(), max_value
            )
        )
    if inplace:
        array[:] = array & ~mask
        array[:] = array | ((sub_field_array << lsb) & mask).astype(array.dtype)
    else:
        array = array & ~mask
        return array | ((sub_field_array << lsb) & mask).astype(array.dtype)



from .base import LasBase
from ..laswriter import LasWriter
from ..utils import ctypes_max_limit
from ..compression import LazBackend


class LasData(LasBase):
    def __init__(self, *, header=None, vlrs=None, points=None, evlrs=None):
        super().__init__(header=header, vlrs=vlrs, points=points)
        self.laszip_was_called = False
        self.evlrs = [] if evlrs is None else evlrs

    def update_header(self):
        super().update_header()
        if len(self.vlrs.get("WktCoordinateSystemVlr")) == 1:
            self.header.global_encoding.wkt = 1

        self.header.start_of_waveform_data_packet_record = 0

        if len(self.points) > ctypes_max_limit(
            self.header.__class__.legacy_point_count.size
        ):
            self.header.legacy_point_count = 0
        else:
            self.header.legacy_point_count = len(self.points)

    def write_to(
        self, out_stream, do_compress=False, laz_backend=LazBackend.detect_available()
    ):
        with LasWriter(
            out_stream,
            self.header,
            self.vlrs,
            do_compress=do_compress,
            closefd=False,
            laz_backend=laz_backend,
        ) as writer:
            writer.write(self.points)
            writer.write_evlrs(self.evlrs)



import ctypes
import datetime
import enum
import logging
import uuid
from typing import Union, BinaryIO

import numpy as np

from .. import compression, utils
from .. import errors
from ..point.record import PointRecord

logger = logging.getLogger(__name__)

LAS_FILE_SIGNATURE = b"LASF"
PROJECT_NAME = b"pylas"


class GpsTimeType(enum.IntEnum):
    WEEK_TIME = 0
    STANDARD = 1


class GlobalEncoding(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("_gps_time_type", ctypes.c_uint16, 1),
        ("waveform_internal", ctypes.c_uint16, 1),  # 1.3
        ("waveform_external", ctypes.c_uint16, 1),  # 1.3
        ("synthetic_return_numbers", ctypes.c_uint16, 1),  # 1.3
        ("wkt", ctypes.c_uint16, 1),  # 1.4
        ("reserved", ctypes.c_uint16, 11),
    ]

    def are_waveform_flag_equal(self):
        return self.waveform_internal == self.waveform_external

    @property
    def gps_time_type(self):
        if self._gps_time_type:
            return GpsTimeType.STANDARD
        else:
            return GpsTimeType.WEEK_TIME

    @gps_time_type.setter
    def gps_time_type(self, value):
        self._gps_time_type = value


class RawHeader1_1(ctypes.LittleEndianStructure):
    _version_ = "1.1"
    _pack_ = 1
    _fields_ = [
        ("file_signature", ctypes.c_char * 4),
        ("file_source_id", ctypes.c_uint16),
        ("global_encoding", GlobalEncoding),
        ("uuid_bytes", ctypes.c_ubyte * 16),
        ("version_major", ctypes.c_uint8),
        ("version_minor", ctypes.c_uint8),
        ("system_identifier", ctypes.c_char * 32),
        ("generating_software", ctypes.c_char * 32),
        ("creation_day_of_year", ctypes.c_uint16),
        ("creation_year", ctypes.c_uint16),
        ("size", ctypes.c_uint16),
        ("offset_to_point_data", ctypes.c_uint32),
        ("number_of_vlr", ctypes.c_uint32),
        ("_point_data_format_id", ctypes.c_uint8),
        ("point_data_record_length", ctypes.c_uint16),
        ("legacy_point_count", ctypes.c_uint32),
        ("legacy_number_of_points_by_return", ctypes.c_uint32 * 5),
        ("x_scale", ctypes.c_double),
        ("y_scale", ctypes.c_double),
        ("z_scale", ctypes.c_double),
        ("x_offset", ctypes.c_double),
        ("y_offset", ctypes.c_double),
        ("z_offset", ctypes.c_double),
        ("x_max", ctypes.c_double),
        ("x_min", ctypes.c_double),
        ("y_max", ctypes.c_double),
        ("y_min", ctypes.c_double),
        ("z_max", ctypes.c_double),
        ("z_min", ctypes.c_double),
    ]

    def __init__(self):
        super().__init__(
            file_signature=LAS_FILE_SIGNATURE,
            version=self._version_,
            generating_software=PROJECT_NAME,
            size=LAS_HEADERS_SIZE[self._version_],
            offset_to_point_data=LAS_HEADERS_SIZE[self._version_],
            x_scale=0.001,
            y_scale=0.001,
            z_scale=0.001,
        )
        self.date = datetime.date.today()

    @property
    def point_count(self):
        """Returns the number of points in the file"""
        return self.legacy_point_count

    @point_count.setter
    def point_count(self, value):
        self.legacy_point_count = value

    @property
    def number_of_points_by_return(self):
        return self.legacy_number_of_points_by_return

    @number_of_points_by_return.setter
    def number_of_points_by_return(self, value):
        if len(value) > 5:
            logger.warning(
                "Received return numbers up to {}, truncating to 5 for header.".format(
                    len(value)
                )
            )
        self.legacy_number_of_points_by_return = tuple(value[:5])

    @property
    def version(self):
        """
        .. DANGER::
            You should not use this directly if you want to change/convert
            the header version as this function does not change the underlying
            size of bytes buffer.
            See :meth:`HeaderFactory.convert_header` to convert headers

        """
        return "{}.{}".format(self.version_major, self.version_minor)

    @version.setter
    def version(self, new_version):
        try:
            self.size = LAS_HEADERS_SIZE[str(new_version)]
            self.offset_to_point_data = self.size
        except KeyError:
            raise errors.FileVersionNotSupported(new_version)
        self.version_major, self.version_minor = map(int, new_version.split("."))

    @property
    def date(self):
        """Returns the creation date stored in the las file

        Returns
        -------
        datetime.date

        """
        try:
            return datetime.date(self.creation_year, 1, 1) + datetime.timedelta(
                self.creation_day_of_year - 1
            )
        except ValueError:
            return None

    @date.setter
    def date(self, date):
        """Returns the date of file creation as a python date object"""
        self.creation_year = date.year
        self.creation_day_of_year = date.timetuple().tm_yday

    @property
    def point_format_id(self):
        return compression.compressed_id_to_uncompressed(self._point_data_format_id)

    @point_format_id.setter
    def point_format_id(self, value):
        """Returns the point format id of the points"""
        self._point_data_format_id = value

    @property
    def point_size(self):
        """Returns the number of bits each point takes"""
        return self.point_data_record_length

    @point_size.setter
    def point_size(self, value):
        self.point_data_record_length = value

    @property
    def uuid(self):
        return uuid.UUID(bytes_le=bytes(self.uuid_bytes))

    @uuid.setter
    def uuid(self, new_uuid):
        self.uuid_bytes = (ctypes.c_ubyte * 16)(*new_uuid.bytes_le)

    @property
    def are_points_compressed(self):
        """Returns True if the point_format_id indicates that the points are stored compressed"""
        return compression.is_point_format_compressed(self._point_data_format_id)

    @property
    def mins(self):
        """Returns de minimum values of x, y, z as a numpy array"""
        return np.array([self.x_min, self.y_min, self.z_min])

    @mins.setter
    def mins(self, value):
        """Sets de minimum values of x, y, z as a numpy array"""
        self.x_min, self.y_min, self.z_min = value

    @property
    def maxs(self):
        """Returns de maximum values of x, y, z as a numpy array"""
        return np.array([self.x_max, self.y_max, self.z_max])

    @maxs.setter
    def maxs(self, value):
        """Sets de maximum values of x, y, z as a numpy array"""
        self.x_max, self.y_max, self.z_max = value

    @property
    def scales(self):
        """Returns the scaling values of x, y, z as a numpy array"""
        return np.array([self.x_scale, self.y_scale, self.z_scale])

    @scales.setter
    def scales(self, value):
        self.x_scale, self.y_scale, self.z_scale = value

    @property
    def offsets(self):
        """Returns the offsets values of x, y, z as a numpy array"""
        return np.array([self.x_offset, self.y_offset, self.z_offset])

    @offsets.setter
    def offsets(self, value):
        self.x_offset, self.y_offset, self.z_offset = value

    def write_to(self, out_stream):
        hdr_bytes = bytearray(self)

        slc = hdr_bytes[26 : 26 + 32]
        i = slc.find(b"\0")
        slc[i:] = b"\x00"

        slc = hdr_bytes[58 : 58 + 32]
        i = slc.find(b"\0")
        slc[i:] = b"\x00"

        out_stream.write(hdr_bytes)

    def partial_reset(self):
        self.generating_software = PROJECT_NAME
        self.date = datetime.date.today()
        self.offset_to_point_data = LAS_HEADERS_SIZE[self._version_]
        self.point_count = 0
        self.number_of_vlr = 0
        self.number_of_points_by_return = (0, 0, 0, 0, 0)
        self.mins = [0.0, 0.0, 0.0]
        self.maxs = [0.0, 0.0, 0.0]

    def set_point_count_to_max(self):
        self.point_count = np.iinfo(np.uint32).max

    def set_compressed(self, compressed: bool):
        if compressed:
            self._point_data_format_id = compression.uncompressed_id_to_compressed(
                self._point_data_format_id
            )
        else:
            self._point_data_format_id = compression.compressed_id_to_uncompressed(
                self._point_data_format_id
            )

    def update(self, points: PointRecord) -> None:
        self.x_max = max(
            self.x_max,
            (points["X"].max() * self.x_scale) + self.x_offset,
        )
        self.y_max = max(
            self.y_max,
            (points["Y"].max() * self.y_scale) + self.y_offset,
        )
        self.z_max = max(
            self.z_max,
            (points["Z"].max() * self.z_scale) + self.z_offset,
        )
        self.x_min = min(
            self.x_min,
            (points["X"].min() * self.x_scale) + self.x_offset,
        )
        self.y_min = min(
            self.y_min,
            (points["Y"].min() * self.y_scale) + self.y_offset,
        )
        self.z_min = min(
            self.z_min,
            (points["Z"].min() * self.z_scale) + self.z_offset,
        )

        for return_number, count in zip(*np.unique(points.return_number, return_counts=True)):
            if return_number == 0:
                continue
            if return_number > len(self.number_of_points_by_return):
                break  # np.unique sorts unique values
            self.number_of_points_by_return[return_number - 1] += count
        self.point_count += len(points)

    def __repr__(self):
        return "<LasHeader({})>".format(self.version)


class RawHeader1_2(RawHeader1_1):
    _pack_ = 1
    _version_ = "1.2"


class RawHeader1_3(RawHeader1_2):
    _pack_ = 1
    _version_ = "1.3"
    _fields_ = [("start_of_waveform_data_packet_record", ctypes.c_uint64)]


class RawHeader1_4(RawHeader1_3):
    _pack_ = 1
    _version_ = "1.4"
    _fields_ = [
        ("start_of_first_evlr", ctypes.c_uint64),
        ("number_of_evlr", ctypes.c_uint32),
        ("_point_count", ctypes.c_uint64),
        ("_number_of_points_by_return", ctypes.c_uint64 * 15),
    ]

    @property
    def point_count(self):
        return self._point_count

    @point_count.setter
    def point_count(self, value):
        if value > utils.ctypes_max_limit(ctypes.sizeof(ctypes.c_uint32)):
            self.legacy_point_count = 0
        else:
            self.legacy_point_count = value
        self._point_count = value

    @property
    def number_of_points_by_return(self):
        return self._number_of_points_by_return

    @number_of_points_by_return.setter
    def number_of_points_by_return(self, value):
        value = tuple(value)
        if len(value) > 15:
            logger.warning(
                "Received return numbers up to {}, truncating to 15 for header.".format(
                    len(value)
                )
            )
            self.legacy_number_of_points_by_return = [0] * 5
        else:
            self.legacy_number_of_points_by_return = value[:5]

        self._number_of_points_by_return = value[:15]

    def update_evlrs_info_in_stream(self, stream, start=0):
        current_pos = stream.tell()
        stream.seek(start + RawHeader1_4.start_of_first_evlr.offset)
        stream.write(self.start_of_first_evlr.to_bytes(8, "little"))
        stream.write(self.number_of_evlr.to_bytes(4, "little"))
        stream.seek(current_pos)

    def set_point_count_to_max(self):
        super().set_point_count_to_max()
        self._point_count = np.iinfo(np.uint64).max

    def partial_reset(self):
        super().partial_reset()
        self.start_of_first_evlr = 0
        self.number_of_evlr = 0
        self.point_count = 0
        self.number_of_points_by_return = [0] * 15


Header = Union[
    RawHeader1_1,
    RawHeader1_2,
    RawHeader1_3,
    RawHeader1_4,
]


class HeaderFactory:
    """Factory to create a new header by specifying the version.
    This Factory also handles converting headers between different
    versions.
    """

    _version_to_header = {
        "1.1": RawHeader1_1,
        "1.2": RawHeader1_2,
        "1.3": RawHeader1_3,
        "1.4": RawHeader1_4,
    }
    _offset_to_major_version = RawHeader1_1.version_major.offset

    @classmethod
    def header_class_for_version(cls, version: Union[str, float]):
        """
        >>> HeaderFactory.header_class_for_version(2.0)
        Traceback (most recent call last):
         ...
        pylas.errors.FileVersionNotSupported: 2.0

        >>> HeaderFactory.header_class_for_version(1.2)
        <class 'pylas.headers.rawheader.RawHeader1_2'>

        >>> header_class = HeaderFactory.header_class_for_version(1.4)
        >>> header_class()
        <LasHeader(1.4)>

        """
        try:
            return cls._version_to_header[str(version)]
        except KeyError:
            raise errors.FileVersionNotSupported(version)

    @classmethod
    def new(cls, version: Union[str, float]) -> Header:
        """Returns a new instance of a header.

        Parameters
        ----------
        version : float or str
            The header version


        >>> HeaderFactory.new(1.4)
        <LasHeader(1.4)>

        >>> HeaderFactory.new('1.2')
        <LasHeader(1.2)>

        """
        return cls.header_class_for_version(version)()

    @classmethod
    def read_from_stream(cls, stream: BinaryIO) -> Header:
        sizeof_u8 = ctypes.sizeof(ctypes.c_uint8)
        header_bytes = bytearray(
            stream.read(cls._offset_to_major_version + (sizeof_u8 * 2))
        )

        if header_bytes[:4] != LAS_FILE_SIGNATURE:
            raise errors.PylasError(
                "File Signature ({}) is not {}".format(
                    header_bytes[:4], LAS_FILE_SIGNATURE
                )
            )

        major = ctypes.c_uint8.from_buffer(
            header_bytes, cls._offset_to_major_version
        ).value
        minor = ctypes.c_uint8.from_buffer(
            header_bytes, cls._offset_to_major_version + sizeof_u8
        ).value
        version = "{}.{}".format(major, minor)

        header_class = cls.header_class_for_version(version)
        header_bytes += stream.read(ctypes.sizeof(header_class) - len(header_bytes))
        return header_class.from_buffer(header_bytes)

    @classmethod
    def from_mmap(cls, mmap):
        version = cls.peek_file_version(mmap)
        return cls.header_class_for_version(version).from_buffer(mmap)

    @classmethod
    def peek_file_version(cls, stream: BinaryIO) -> str:
        """seeks to the position of the las version header fields
        in the stream and returns it as a str

        Parameters
        ----------
        stream io.BytesIO

        Returns
        -------
        str
            file version read from the stream

        """
        old_pos = stream.tell()
        stream.seek(cls._offset_to_major_version)
        major = int.from_bytes(stream.read(ctypes.sizeof(ctypes.c_uint8)), "little")
        minor = int.from_bytes(stream.read(ctypes.sizeof(ctypes.c_uint8)), "little")
        stream.seek(old_pos)
        return f"{major}.{minor}"

    @classmethod
    def convert_header(
        cls, old_header: Header, new_version: Union[str, float]
    ) -> Header:
        """Converts a header to a another version

        Parameters
        ----------
        old_header: the old header instance
        new_version: float or str

        Returns
        -------
        The converted header


        >>> old_header = HeaderFactory.new(1.2)
        >>> HeaderFactory.convert_header(old_header, 1.4)
        <LasHeader(1.4)>

        >>> old_header = HeaderFactory.new('1.4')
        >>> HeaderFactory.convert_header(old_header, '1.2')
        <LasHeader(1.2)>

        """
        new_header_class = cls.header_class_for_version(new_version)

        b = bytearray(old_header)
        b += b"\x00" * (ctypes.sizeof(new_header_class) - len(b))
        new_header = new_header_class.from_buffer(b)
        new_header.version = str(new_version)

        return new_header


LAS_HEADERS_SIZE = {
    "1.1": ctypes.sizeof(RawHeader1_1),
    "1.2": ctypes.sizeof(RawHeader1_1),
    "1.3": ctypes.sizeof(RawHeader1_3),
    "1.4": ctypes.sizeof(RawHeader1_4),
}



import logging
from typing import BinaryIO, Type, List

from .known import vlr_factory, IKnownVLR
from .rawvlr import RawVLR

logger = logging.getLogger(__name__)


class RawVLRList:
    """A RawVLRList is like a VLR list but it should only
    hold RawVLRs.

    This class is meant to make it easier to write VLRS the the file and know in advance
    the size in bytes taken by all the VLRs combined

    """

    def __init__(self, iterable=None):
        if iterable is not None:
            self.vlrs = list(iterable)
        else:
            self.vlrs = []

    def append(self, raw_vlr):
        self.vlrs.append(raw_vlr)

    def __len__(self):
        return len(self.vlrs)

    def total_size_in_bytes(self):
        return sum(v.size_in_bytes() for v in self.vlrs)

    def write_to(self, out_stream):
        """Writes all the raw vlrs contained in list to
        the out_stream

        Parameters
        ----------
        out_stream: io.RawIOBase
            The stream where vlrs will be written to

        """
        for vlr in self.vlrs:
            vlr.write_to(out_stream)

    def __iter__(self):
        return iter(self.vlrs)

    @classmethod
    def from_list(cls, vlrs):
        """Construct a RawVLR list from a list of vlrs

        Parameters
        ----------
        vlrs: iterable of VLR

        Returns
        -------
        RawVLRList

        """
        raw_vlrs = cls()
        for vlr in vlrs:
            raw = RawVLR()
            raw.header.user_id = vlr.user_id.encode("utf8")
            raw.header.description = vlr.description.encode("utf8")
            raw.header.record_id = vlr.record_id
            raw.record_data = vlr.record_data_bytes()
            raw_vlrs.append(raw)
        return raw_vlrs


class VLRList:
    """Class responsible for managing the vlrs"""

    def __init__(self):
        self.vlrs = []

    def append(self, vlr):
        """append a vlr to the list

        Parameters
        ----------
        vlr: RawVlR or VLR or KnownVlr

        Returns
        -------

        """
        self.vlrs.append(vlr)

    def extend(self, vlr_list):
        """append all elements of the vlr_list into self"""
        self.vlrs.extend(vlr_list)

    def get_by_id(self, user_id="", record_ids=(None,)):
        """Function to get vlrs by user_id and/or record_ids.
        Always returns a list even if only one vlr matches the user_id and record_id

        >>> import pylas
        >>> from pylas.vlrs.known import ExtraBytesVlr, WktCoordinateSystemVlr
        >>> las = pylas.read("pylastests/extrabytes.las")
        >>> las.vlrs
        [<ExtraBytesVlr(extra bytes structs: 5)>]
        >>> las.vlrs.get(WktCoordinateSystemVlr.official_user_id())
        []
        >>> las.vlrs.get(WktCoordinateSystemVlr.official_user_id())[0]
        Traceback (most recent call last):
        IndexError: list index out of range
        >>> las.vlrs.get_by_id(ExtraBytesVlr.official_user_id())
        [<ExtraBytesVlr(extra bytes structs: 5)>]
        >>> las.vlrs.get_by_id(ExtraBytesVlr.official_user_id())[0]
        <ExtraBytesVlr(extra bytes structs: 5)>

        Parameters
        ----------
        user_id: str, optional
                 the user id
        record_ids: iterable of int, optional
                    THe record ids of the vlr(s) you wish to get

        Returns
        -------
        :py:class:`list`
            a list of vlrs matching the user_id and records_ids

        """
        if user_id != "" and record_ids != (None,):
            return [
                vlr
                for vlr in self.vlrs
                if vlr.user_id == user_id and vlr.record_id in record_ids
            ]
        else:
            return [
                vlr
                for vlr in self.vlrs
                if vlr.user_id == user_id or vlr.record_id in record_ids
            ]

    def get(self, vlr_type):
        """Returns the list of vlrs of the requested type
        Always returns a list even if there is only one VLR of type vlr_type.

        >>> import pylas
        >>> las = pylas.read("pylastests/extrabytes.las")
        >>> las.vlrs
        [<ExtraBytesVlr(extra bytes structs: 5)>]
        >>> las.vlrs.get("WktCoordinateSystemVlr")
        []
        >>> las.vlrs.get("WktCoordinateSystemVlr")[0]
        Traceback (most recent call last):
        IndexError: list index out of range
        >>> las.vlrs.get('ExtraBytesVlr')
        [<ExtraBytesVlr(extra bytes structs: 5)>]
        >>> las.vlrs.get('ExtraBytesVlr')[0]
        <ExtraBytesVlr(extra bytes structs: 5)>


        Parameters
        ----------
        vlr_type: str
                  the class name of the vlr

        Returns
        -------
        :py:class:`list`
            a List of vlrs matching the user_id and records_ids

        """
        return [v for v in self.vlrs if v.__class__.__name__ == vlr_type]

    def extract(self, vlr_type: Type[IKnownVLR]) -> List[IKnownVLR]:
        """Returns the list of vlrs of the requested type
        The difference with get is that the returned vlrs will be removed from the list

        Parameters
        ----------
        vlr_type: str
                  the class name of the vlr

        Returns
        -------
        list
            a List of vlrs matching the user_id and records_ids

        """
        kept_vlrs, extracted_vlrs = [], []
        for vlr in self.vlrs:
            if vlr.__class__.__name__ == vlr_type:
                extracted_vlrs.append(vlr)
            else:
                kept_vlrs.append(vlr)
        self.vlrs = kept_vlrs
        return extracted_vlrs

    def pop(self, index):
        return self.vlrs.pop(index)

    def index(self, vlr_type):
        for i, v in enumerate(self.vlrs):
            if v.__class__.__name__ == vlr_type:
                return i
        raise ValueError("{} is not in the VLR list".format(vlr_type))

    def __iter__(self):
        yield from iter(self.vlrs)

    def __getitem__(self, item):
        return self.vlrs[item]

    def __len__(self):
        return len(self.vlrs)

    def __eq__(self, other):
        if isinstance(other, list):
            return self.vlrs == other

    def __repr__(self):
        return "[{}]".format(", ".join(repr(vlr) for vlr in self.vlrs))

    @classmethod
    def read_from(cls, data_stream: BinaryIO, num_to_read: int) -> 'VLRList':
        """Reads vlrs and parse them if possible from the stream

        Parameters
        ----------
        data_stream : io.BytesIO
                      stream to read from
        num_to_read : int
                      number of vlrs to be read

        Returns
        -------
        pylas.vlrs.vlrlist.VLRList
            List of vlrs

        """
        vlrlist = cls()
        for _ in range(num_to_read):
            raw = RawVLR.read_from(data_stream)
            vlrlist.append(vlr_factory(raw))

        return vlrlist

    @classmethod
    def from_list(cls, vlr_list):
        vlrs = cls()
        vlrs.vlrs = vlr_list
        return vlrs



from setuptools import setup, find_packages

with open("README.rst") as f:
    readme = f.read()

setup(
    name="pylas",
    version="0.5.0a3",
    description="Las/Laz reading and writing in python",
    long_description=readme,
    url="https://github.com/tmontaigu/pylas",
    author="Thomas Montaigu",
    author_email="thomas.montaigu@laposte.net",
    python_requires=">=3.6",
    keywords="las lidar",
    license="BSD 3-Clause",
    packages=find_packages(exclude=("pylastests",)),
    zip_safe=False,
    install_requires=["numpy"],
    extras_require={
        "dev": [
            "pytest",
            "sphinx",
            "sphinx-rtd-theme"
        ],
        "lazrs": [
            "lazrs>=0.2.0, < 0.3.0"
        ]
    }
)



import abc
import io
import logging
import subprocess
from copy import copy
from typing import BinaryIO, Optional, Union, Iterable

import numpy as np

from .compression import LazBackend
from .compression import find_laszip_executable
from .errors import PylasError, LazError
from .evlrs import EVLRList, RawEVLRList
from .headers import HeaderFactory
from .headers.rawheader import Header
from .point import dims
from .point.format import PointFormat
from .point.record import PointRecord
from .utils import ConveyorThread
from .vlrs.known import LasZipVlr, ExtraBytesStruct, ExtraBytesVlr
from .vlrs.vlrlist import VLRList, RawVLRList

logger = logging.getLogger(__name__)

try:
    import lazrs
except ModuleNotFoundError:
    pass


class PointWriter(abc.ABC):
    @property
    @abc.abstractmethod
    def destination(self):
        ...

    def write_initial_header_and_vlrs(self, header, vlrs: VLRList):
        raw_vlrs = RawVLRList.from_list(vlrs)
        header.number_of_vlr = len(raw_vlrs)
        header.offset_to_point_data = header.size + raw_vlrs.total_size_in_bytes()

        header.write_to(self.destination)
        raw_vlrs.write_to(self.destination)

    @abc.abstractmethod
    def write_points(self, points):
        ...

    @abc.abstractmethod
    def done(self):
        ...

    def write_updated_header(self, header):
        self.destination.seek(0, io.SEEK_SET)
        header.write_to(self.destination)


class LasWriter:
    def __init__(
        self,
        dest: BinaryIO,
        header: Header,
        vlrs: Optional[VLRList] = None,
        do_compress: bool = False,
        laz_backend: Union[
            LazBackend, Iterable[LazBackend]
        ] = LazBackend.detect_available(),
        closefd: bool = True,
    ) -> None:
        self.closefd = closefd
        self.header = copy(header)
        self.header.partial_reset()
        self.header.maxs = [np.finfo("f8").min] * 3
        self.header.mins = [np.finfo("f8").max] * 3

        self.do_compress = do_compress
        self.laz_backend = laz_backend
        self.dest = dest

        if vlrs is None:
            self.vlrs = VLRList()
        else:
            self.vlrs = vlrs

        # These will be initialized on the first call to `write`
        self.point_format: PointFormat
        self.point_writer: PointWriter
        self.done = False

    def write(self, points: PointRecord) -> None:
        if not points:
            return

        if self.done:
            raise PylasError("Cannot write points anymore")

        if self.header.point_count == 0:
            dims.raise_if_version_not_compatible_with_fmt(
                points.point_format.id, self.header.version
            )
            self.point_format = points.point_format
            self.header.point_format_id = self.point_format.id
            self.header.point_size = self.point_format.size
            self.header.set_compressed(self.do_compress)

            if self.point_format.num_extra_bytes > 0:
                try:
                    self.vlrs.index("ExtraBytesVlr")
                except ValueError:
                    extra_bytes_vlr = ExtraBytesVlr()
                    for dim_info in self.point_format.extra_dimensions:
                        extra_byte = ExtraBytesStruct.from_dimension_info(dim_info)
                        extra_bytes_vlr.extra_bytes_structs.append(extra_byte)
                    self.vlrs.append(extra_bytes_vlr)

            if self.do_compress:
                self.point_writer = self._create_laz_backend(self.laz_backend)
            else:
                self.point_writer = UncompressedPointWriter(self.dest)

            self.point_writer.write_initial_header_and_vlrs(self.header, self.vlrs)
        elif points.point_format != self.point_format:
            raise PylasError("Incompatible point formats")

        self.header.update(points)
        self.point_writer.write_points(points)

    def write_evlrs(self, evlrs: EVLRList) -> None:
        if self.header.version < "1.4":
            raise PylasError(
                "EVLRs are not supported on files with version less than 1.4"
            )

        if len(evlrs) > 0:
            self.point_writer.done()
            self.done = True
            self.header.number_of_evlr = len(evlrs)
            self.header.start_of_first_evlr = self.dest.tell()
            # self.header.update_evlrs_info_in_stream(self.dest)
            raw_evlrs = RawEVLRList.from_list(evlrs)
            raw_evlrs.write_to(self.dest)

    def close(self) -> None:
        if self.point_writer is not None:
            if not self.done:
                self.point_writer.done()
            self.point_writer.write_updated_header(self.header)
        if self.closefd:
            self.dest.close()

    def _create_laz_backend(
        self, laz_backends: Union[LazBackend, Iterable[LazBackend]]
    ) -> PointWriter:
        try:
            laz_backends = iter(laz_backends)
        except TypeError:
            laz_backends = (laz_backends,)

        last_error: Optional[Exception] = None
        for backend in laz_backends:
            try:
                if not backend.is_available():
                    raise PylasError(f"The '{backend}' is not available")

                if backend == LazBackend.Laszip:
                    return LasZipProcessPointWriter(self.dest)
                elif backend == LazBackend.LazrsParallel:
                    return LazrsPointWriter(self.dest, self.point_format, parallel=True)
                elif backend == LazBackend.Lazrs:
                    return LazrsPointWriter(
                        self.dest, self.point_format, parallel=False
                    )
                else:
                    raise PylasError("Unknown LazBacked: {}".format(backend))
            except Exception as e:
                logger.error(e)
                last_error = e

        if last_error is not None:
            raise PylasError("No LazBackend selected, cannot compress")
        else:
            raise PylasError(f"No LazBackend could be initialized: {last_error}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class UncompressedPointWriter(PointWriter):
    def __init__(self, dest):
        self.dest = dest

    @property
    def destination(self):
        return self.dest

    def write_points(self, points):
        self.dest.write(points.memoryview())

    def done(self):
        pass


class LasZipProcessPointWriter(PointWriter):
    def __init__(self, dest):
        laszip_binary = find_laszip_executable()
        self.dest = dest

        self.conveyor: Optional[ConveyorThread]
        try:
            _ = dest.fileno()
        except OSError:
            # FIXME
            # For some reason, when piping to stdout, creates files
            # that are missing the chunk table
            self.process = subprocess.Popen(
                [laszip_binary, "-stdin", "-olaz", "-stdout"],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            self.conveyor = ConveyorThread(self.process.stdout, self.dest)
            self.conveyor.start()
        else:
            self.conveyor = None
            self.process = subprocess.Popen(
                [laszip_binary, "-stdin", "-olaz", "-stdout"],
                stdin=subprocess.PIPE,
                stdout=self.dest,
                stderr=subprocess.PIPE,
            )

    @property
    def destination(self):
        return self.process.stdin

    def write_initial_header_and_vlrs(self, header, vlrs):
        # If the header we sent to laszip has a point_count
        # that is less than the number of points that will get
        # pass to calls to write_points, the laszip process
        # will quit with an error, that's why we set it to max
        header.set_point_count_to_max()
        header.set_compressed(False)
        super(LasZipProcessPointWriter, self).write_initial_header_and_vlrs(
            header, vlrs
        )
        header.set_compressed(True)
        header.point_count = 0

    def write_points(self, points):
        if self.process.poll() is not None:
            self.raise_if_bad_err_code()
        else:
            try:
                self.process.stdin.write(points.memoryview())
            except BrokenPipeError:
                raise LazError(
                    "Laszip process failed: {}".format(
                        self.process.stderr.read().decode()
                    )
                ) from None

    def done(self):
        self.process.stdin.flush()
        self.process.stdin.close()
        if self.conveyor is not None:
            self.conveyor.join()

        self.process.wait()
        self.process.poll()
        self.raise_if_bad_err_code()
        self._rewrite_chunk_table_offset()

    def _rewrite_chunk_table_offset(self):
        position_backup = self.dest.tell()
        self.dest.seek(0, io.SEEK_SET)
        hdr = HeaderFactory.read_from_stream(self.dest)
        self.dest.seek(hdr.offset_to_point_data, io.SEEK_SET)
        offset_to_chunk_table = int.from_bytes(self.dest.read(8), "little", signed=True)
        if offset_to_chunk_table == -1:
            self.dest.seek(-8, io.SEEK_END)
            offset_to_chunk_table = int.from_bytes(
                self.dest.read(8), "little", signed=True
            )
            self.dest.seek(hdr.offset_to_point_data, io.SEEK_SET)
            self.dest.write(offset_to_chunk_table.to_bytes(8, "little", signed=True))
            self.dest.seek(-8, io.SEEK_END)
            self.dest.truncate()
        else:
            self.dest.seek(position_backup, io.SEEK_SET)

    def write_updated_header(self, header):
        position_save = self.dest.tell()
        self.dest.seek(0, io.SEEK_SET)
        hdr = HeaderFactory.read_from_stream(self.dest)
        hdr.point_count = header.point_count
        hdr.maxs = header.maxs
        hdr.mins = header.mins
        hdr.number_of_points_by_return = header.number_of_points_by_return

        if header.version >= "1.4":
            hdr.number_of_evlr = header.number_of_evlr
            hdr.start_of_first_evlr = header.start_of_first_evlr

        self.dest.seek(0, io.SEEK_SET)
        hdr.write_to(self.dest)
        self.dest.seek(position_save, io.SEEK_SET)

    def raise_if_bad_err_code(self):
        if self.process.returncode != 0:
            error_msg = self.process.stderr.read().decode()
            raise LazError(
                "Laszip failed to {} with error code {}\n\t{}".format(
                    "compress",
                    self.process.returncode,
                    "\n\t".join(error_msg.splitlines()),
                )
            )


class LazrsPointWriter(PointWriter):
    def __init__(self, dest: BinaryIO, point_format: PointFormat, parallel: bool):
        self.dest = dest
        self.vlr = lazrs.LazVlr.new_for_compression(
            point_format.id, point_format.num_extra_bytes
        )
        self.parallel = parallel
        self.compressor = None

    def write_initial_header_and_vlrs(self, header, vlrs: VLRList):
        laszip_vlr = LasZipVlr(self.vlr.record_data())
        vlrs.append(laszip_vlr)
        super().write_initial_header_and_vlrs(header, vlrs)
        # We have to initialize our compressor here
        # because on init, it writes the offset to chunk table
        # so the header and vlrs have to be written
        if self.parallel:
            self.compressor = lazrs.ParLasZipCompressor(self.dest, self.vlr)
        else:
            self.compressor = lazrs.LasZipCompressor(self.dest, self.vlr)

    @property
    def destination(self):
        return self.dest

    def write_points(self, points):
        points_bytes = np.frombuffer(points.array, np.uint8)
        self.compressor.compress_many(points_bytes)

    def done(self):
        if self.compressor is not None:
            self.compressor.done()



"""  This module contains things like the definitions of the point formats dimensions,
the mapping between dimension names and their type, mapping between point format and
compatible file version
"""
import itertools
import operator
from collections import UserDict
from enum import Enum
from typing import (
    NamedTuple,
    Optional,
    Dict,
    Tuple,
    Set,
    Iterable,
    Mapping,
    TypeVar,
    Generic,
    List,
)

import numpy as np

from . import packing
from .. import errors

ValueType = TypeVar("ValueType")


class PointFormatDict(UserDict, Generic[ValueType]):
    """Simple wrapper around a dict that changes
    the exception raised when accessing a key that is not-present

    """

    def __init__(self, wrapped_dict: Dict[int, ValueType]):
        super().__init__(wrapped_dict)

    def __getitem__(self, key: int) -> ValueType:
        try:
            return self.data[key]
        except KeyError:
            raise errors.PointFormatNotSupported(key) from None


class SubField(NamedTuple):
    name: str
    mask: int


def _point_format_to_dtype(
    point_format: Iterable[str], dimensions_to_type: Mapping[str, str]
) -> np.dtype:
    """build the numpy.dtype for a point format

    Parameters:
    ----------
    point_format : iterable of str
        The dimensions names of the point format
    dimensions : dict
        The dictionary of dimensions
    Returns
    -------
    numpy.dtype
        The dtype for the input point format
    """
    return np.dtype(
        [(dim_name, dimensions_to_type[dim_name]) for dim_name in point_format]
    )


def _build_point_formats_dtypes(
    point_format_dimensions: Mapping[int, Tuple[str]],
    dimensions_dict: Mapping[str, str],
) -> Dict[int, np.dtype]:
    """Builds the dict mapping point format id to numpy.dtype
    In the dtypes, bit fields are still packed, and need to be unpacked each time
    you want to access them
    """
    return {
        fmt_id: _point_format_to_dtype(point_fmt, dimensions_dict)
        for fmt_id, point_fmt in point_format_dimensions.items()
    }


# Definition of the points dimensions and formats
# LAS version [1.0, 1.1, 1.2, 1.3, 1.4]
DIMENSIONS_TO_TYPE: Dict[str, str] = {
    "X": "i4",
    "Y": "i4",
    "Z": "i4",
    "intensity": "u2",
    "bit_fields": "u1",
    "raw_classification": "u1",
    "scan_angle_rank": "i1",
    "user_data": "u1",
    "point_source_id": "u2",
    "gps_time": "f8",
    "red": "u2",
    "green": "u2",
    "blue": "u2",
    # Waveform related dimensions
    "wavepacket_index": "u1",
    "wavepacket_offset": "u8",
    "wavepacket_size": "u4",
    "return_point_wave_location": "u4",
    "x_t": "f4",
    "y_t": "f4",
    "z_t": "f4",
    # Las 1.4
    "classification_flags": "u1",
    "scan_angle": "i2",
    "classification": "u1",
    "nir": "u2",
}

POINT_FORMAT_0: Tuple[str, ...] = (
    "X",
    "Y",
    "Z",
    "intensity",
    "bit_fields",
    "raw_classification",
    "scan_angle_rank",
    "user_data",
    "point_source_id",
)

POINT_FORMAT_6: Tuple[str, ...] = (
    "X",
    "Y",
    "Z",
    "intensity",
    "bit_fields",
    "classification_flags",
    "classification",
    "user_data",
    "scan_angle",
    "point_source_id",
    "gps_time",
)

WAVEFORM_FIELDS_NAMES: Tuple[str, ...] = (
    "wavepacket_index",
    "wavepacket_offset",
    "wavepacket_size",
    "return_point_wave_location",
    "x_t",
    "y_t",
    "z_t",
)

COLOR_FIELDS_NAMES: Tuple[str, ...] = ("red", "green", "blue")

POINT_FORMAT_DIMENSIONS = PointFormatDict(
    {
        0: POINT_FORMAT_0,
        1: POINT_FORMAT_0 + ("gps_time",),
        2: POINT_FORMAT_0 + COLOR_FIELDS_NAMES,
        3: POINT_FORMAT_0 + ("gps_time",) + COLOR_FIELDS_NAMES,
        4: POINT_FORMAT_0 + ("gps_time",) + WAVEFORM_FIELDS_NAMES,
        5: POINT_FORMAT_0 + ("gps_time",) + COLOR_FIELDS_NAMES + WAVEFORM_FIELDS_NAMES,
        6: POINT_FORMAT_6,
        7: POINT_FORMAT_6 + COLOR_FIELDS_NAMES,
        8: POINT_FORMAT_6 + COLOR_FIELDS_NAMES + ("nir",),
        9: POINT_FORMAT_6 + WAVEFORM_FIELDS_NAMES,
        10: POINT_FORMAT_6 + COLOR_FIELDS_NAMES + ("nir",) + WAVEFORM_FIELDS_NAMES,
    }
)

# sub fields of the 'bit_fields' dimension
RETURN_NUMBER_MASK_0 = 0b00000111
NUMBER_OF_RETURNS_MASK_0 = 0b00111000
SCAN_DIRECTION_FLAG_MASK_0 = 0b01000000
EDGE_OF_FLIGHT_LINE_MASK_0 = 0b10000000

# sub fields of the 'raw_classification' dimension
CLASSIFICATION_MASK_0 = 0b00011111
SYNTHETIC_MASK_0 = 0b00100000
KEY_POINT_MASK_0 = 0b01000000
WITHHELD_MASK_0 = 0b10000000

# sub fields of the bit_fields
RETURN_NUMBER_MASK_6 = 0b00001111
NUMBER_OF_RETURNS_MASK_6 = 0b11110000

# sub fields of classification flags
CLASSIFICATION_FLAGS_MASK_6 = 0b00001111

SYNTHETIC_MASK_6 = 0b00000001
KEY_POINT_MASK_6 = 0b00000010
WITHHELD_MASK_6 = 0b00000100
OVERLAP_MASK_6 = 0b00001000
SCANNER_CHANNEL_MASK_6 = 0b00110000
SCAN_DIRECTION_FLAG_MASK_6 = 0b01000000
EDGE_OF_FLIGHT_LINE_MASK_6 = 0b10000000

COMPOSED_FIELDS_0: Dict[str, List[SubField]] = {
    "bit_fields": [
        SubField("return_number", RETURN_NUMBER_MASK_0),
        SubField("number_of_returns", NUMBER_OF_RETURNS_MASK_0),
        SubField("scan_direction_flag", SCAN_DIRECTION_FLAG_MASK_0),
        SubField("edge_of_flight_line", EDGE_OF_FLIGHT_LINE_MASK_0),
    ],
    "raw_classification": [
        SubField("classification", CLASSIFICATION_MASK_0),
        SubField("synthetic", SYNTHETIC_MASK_0),
        SubField("key_point", KEY_POINT_MASK_0),
        SubField("withheld", WITHHELD_MASK_0),
    ],
}


COMPOSED_FIELDS_6: Dict[str, List[SubField]] = {
    "bit_fields": [
        SubField("return_number", RETURN_NUMBER_MASK_6),
        SubField("number_of_returns", NUMBER_OF_RETURNS_MASK_6),
    ],
    "classification_flags": [
        SubField("synthetic", SYNTHETIC_MASK_6),
        SubField("key_point", KEY_POINT_MASK_6),
        SubField("withheld", WITHHELD_MASK_6),
        SubField("overlap", OVERLAP_MASK_6),
        SubField("scanner_channel", SCANNER_CHANNEL_MASK_6),
        SubField("scan_direction_flag", SCAN_DIRECTION_FLAG_MASK_6),
        SubField("edge_of_flight_line", EDGE_OF_FLIGHT_LINE_MASK_6),
    ],
}

# Dict giving the composed fields for each point_format_id
COMPOSED_FIELDS = PointFormatDict(
    {
        0: COMPOSED_FIELDS_0,
        1: COMPOSED_FIELDS_0,
        2: COMPOSED_FIELDS_0,
        3: COMPOSED_FIELDS_0,
        4: COMPOSED_FIELDS_0,
        5: COMPOSED_FIELDS_0,
        6: COMPOSED_FIELDS_6,
        7: COMPOSED_FIELDS_6,
        8: COMPOSED_FIELDS_6,
        9: COMPOSED_FIELDS_6,
        10: COMPOSED_FIELDS_6,
    }
)

VERSION_TO_POINT_FMT: Dict[str, Tuple[int, ...]] = {
    "1.2": (0, 1, 2, 3),
    "1.3": (0, 1, 2, 3, 4, 5),
    "1.4": (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
}

POINT_FORMATS_DTYPE = PointFormatDict(
    _build_point_formats_dtypes(POINT_FORMAT_DIMENSIONS, DIMENSIONS_TO_TYPE)
)
# This Dict maps point_format_ids to their dimensions names
ALL_POINT_FORMATS_DIMENSIONS = PointFormatDict({**POINT_FORMAT_DIMENSIONS})
# This Dict maps point_format_ids to their numpy.dtype
# the dtype corresponds to the de packed data
ALL_POINT_FORMATS_DTYPE = PointFormatDict({**POINT_FORMATS_DTYPE})


def get_sub_fields_dict(point_format_id: int) -> Dict[str, Tuple[str, SubField]]:
    sub_fields_dict = {}
    for composed_dim_name, sub_fields in COMPOSED_FIELDS[point_format_id].items():
        for sub_field in sub_fields:
            sub_fields_dict[sub_field.name] = (composed_dim_name, sub_field)
    return sub_fields_dict


class DimensionKind(Enum):
    SignedInteger = 0
    UnsignedInteger = 1
    FloatingPoint = 2
    BitField = 3

    @classmethod
    def from_letter(cls, letter: str) -> "DimensionKind":
        if letter == "u":
            return cls.UnsignedInteger
        elif letter == "i":
            return cls.SignedInteger
        elif letter == "f":
            return cls.FloatingPoint
        else:
            raise ValueError(f"Unknown type letter '{letter}'")

    def letter(self) -> Optional[str]:
        if self == DimensionKind.UnsignedInteger:
            return "u"
        elif self == DimensionKind.SignedInteger:
            return "i"
        elif self == DimensionKind.FloatingPoint:
            return "f"
        else:
            return None


def num_bit_set(n: int) -> int:
    """Count the number of bits that are set (1) in the number n

    Brian Kernighan's algorithm
    """
    count = 0
    while n != 0:
        count += 1
        n = n & (n - 1)
    return count


class DimensionInfo(NamedTuple):
    """Tuple that contains information of a dimension"""

    name: str
    kind: DimensionKind
    num_bits: int
    num_elements: int = 1
    is_standard: bool = True

    @classmethod
    def from_type_str(
        cls, name: str, type_str: str, is_standard: bool = True
    ) -> "DimensionInfo":
        first_digits = "".join(itertools.takewhile(lambda l: l.isdigit(), type_str))
        if first_digits:
            num_elements = int(first_digits)
            type_str = type_str[len(first_digits):]
        else:
            num_elements = 1

        kind = DimensionKind.from_letter(type_str[0])
        num_bits = int(type_str[1:]) * 8 * num_elements

        return cls(name, kind, num_bits, num_elements, is_standard)

    @classmethod
    def from_bitmask(
        cls, name: str, bit_mask: int, is_standard: bool = False
    ) -> "DimensionInfo":
        kind = DimensionKind.BitField
        bit_size = num_bit_set(bit_mask)
        return cls(name, kind, bit_size, is_standard=is_standard)

    @property
    def num_bytes(self) -> int:
        return int(self.num_bits // 8)

    @property
    def num_bytes_singular_element(self) -> int:
        return int(self.num_bits // (8 * self.num_elements))

    @property
    def max(self):
        if self.kind == DimensionKind.BitField:
            return (2 ** self.num_bits) - 1
        elif self.kind == DimensionKind.FloatingPoint:
            return np.finfo(self.type_str()).max
        else:
            return np.iinfo(self.type_str()).max

    @property
    def min(self):
        if (
            self.kind == DimensionKind.BitField
            or self.kind == DimensionKind.UnsignedInteger
        ):
            return 0
        elif self.kind == DimensionKind.FloatingPoint:
            return np.finfo(self.type_str()).min
        else:
            return np.iinfo(self.type_str()).min

    def type_str(self) -> Optional[str]:
        if self.kind == DimensionKind.BitField:
            return None

        if self.num_elements == 1:
            return f"{self.kind.letter()}{self.num_bytes_singular_element}"
        return (
            f"{self.num_elements}{self.kind.letter()}{self.num_bytes_singular_element}"
        )


def size_of_point_format_id(point_format_id: int) -> int:
    return ALL_POINT_FORMATS_DTYPE[point_format_id].itemsize


def min_file_version_for_point_format(point_format_id: int) -> str:
    """Returns the minimum file version that supports the given point_format_id"""
    for version, point_formats in sorted(VERSION_TO_POINT_FMT.items()):
        if point_format_id in point_formats:
            return version
    raise errors.PointFormatNotSupported(point_format_id)


def supported_versions() -> Set[str]:
    """Returns the set of supported file versions"""
    return set(VERSION_TO_POINT_FMT.keys())


def supported_point_formats() -> Set[int]:
    """Returns a set of all the point formats supported in pylas"""
    return set(POINT_FORMAT_DIMENSIONS.keys())


def is_point_fmt_compatible_with_version(
    point_format_id: int, file_version: str
) -> bool:
    """Returns true if the file version support the point_format_id"""
    try:
        return point_format_id in VERSION_TO_POINT_FMT[str(file_version)]
    except KeyError:
        raise errors.FileVersionNotSupported(file_version)


def raise_if_version_not_compatible_with_fmt(point_format_id: int, file_version: str):
    if not is_point_fmt_compatible_with_version(point_format_id, file_version):
        raise errors.PylasError(
            "Point format {} is not compatible with file version {}".format(
                point_format_id, file_version
            )
        )


class SubFieldView:
    def __init__(self, array: np.ndarray, bit_mask):
        self.array = array
        self.bit_mask = self.array.dtype.type(bit_mask)
        self.lsb = packing.least_significant_bit_set(bit_mask)
        self.max_value_allowed = int(self.bit_mask >> self.lsb)

    def masked_array(self):
        return (self.array & self.bit_mask) >> self.lsb

    def copy(self):
        return SubFieldView(self.array.copy(), int(self.bit_mask))

    def _do_comparison(self, value, comp):
        if isinstance(value, (int, self.array.dtype)):
            if value > self.max_value_allowed:
                return np.zeros_like(self.array, np.bool)
        return comp(self.array & self.bit_mask, value << self.lsb)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        inpts = []
        for inpt in inputs:
            if isinstance(inpt, self.__class__):
                inpts.append(inpt.masked_array())
            else:
                inpts.append(inpt)
        ret = getattr(ufunc, method)(*inpts, **kwargs)
        if ret is not None and isinstance(ret, np.ndarray):
            if ret.dtype == np.bool:
                return ret
            return self.__class__(ret, int(self.bit_mask))
        return ret

    def __array_function__(self, func, types, args, kwargs):
        if func == np.allclose:
            argslist = []
            for i in range(len(args)):
                if isinstance(args[i], SubFieldView):
                    argslist.append(args[i].masked_array())
                else:
                    argslist.append(args[i])
            return np.allclose(*argslist, **kwargs)
        elif func == np.amax or func == np.max:
            return np.max(self.masked_array())
        elif func == np.amin or func == np.min:
            return np.min(self.masked_array())
        else:
            return func(self.masked_array(), *args[1:], **kwargs)

    def __array__(self, **kwargs):
        return self.masked_array()

    def max(self, **unused_kwargs):
        return self.masked_array().max()

    def min(self, **unused_kwargs):
        return self.masked_array().min()

    def __len__(self):
        return len(self.array)

    def __lt__(self, other):
        return self._do_comparison(other, operator.lt)

    def __le__(self, other):
        return self._do_comparison(other, operator.le)

    def __ge__(self, other):
        return self._do_comparison(other, operator.ge)

    def __gt__(self, other):
        return self._do_comparison(other, operator.gt)

    def __eq__(self, other):
        if isinstance(other, SubFieldView):
            return self.bit_mask == other.bit_mask and self.masked_array() == other
        else:
            return self._do_comparison(other, operator.eq)

    def __ne__(self, other):
        if isinstance(other, SubFieldView):
            return self.bit_mask != other.bit_mask and self.masked_array() != other
        else:
            return self._do_comparison(other, operator.ne)

    def __setitem__(self, key, value):
        if np.max(value) > self.max_value_allowed:
            raise OverflowError(
                f"value {np.max(value)} is greater than allowed (max: {self.max_value_allowed})"
            )
        self.array[key] &= ~self.bit_mask
        self.array[key] |= value << self.lsb

    def __getitem__(self, item):
        return SubFieldView(self.array[item], int(self.bit_mask))

    def __repr__(self):
        return f"<SubFieldView({self.masked_array()})>"


class ScaledArrayView:
    def __init__(self, array, scale: float, offset: float) -> None:
        self.array = array
        self.scale = scale
        self.offset = offset

    def scaled_array(self):
        return self._apply_scale(self.array)

    def copy(self):
        return ScaledArrayView(self.array.copy(), self.scale, self.offset)

    def _apply_scale(self, value):
        return (value * self.scale) + self.offset

    def _remove_scale(self, value):
        return np.round((value - self.offset) / self.scale)

    def max(self, **unused_kwargs):
        return self._apply_scale(self.array.max())

    def min(self, **unused_kwargs):
        return self._apply_scale(self.array.min())

    def __array__(self):
        return self.scaled_array()

    def __array_function__(self, func, types, args, kwargs):
        converted_args = []
        for arg in args:
            if isinstance(arg, (tuple, list)):
                top_level_args = []
                converted_args.append(top_level_args)
            else:
                top_level_args = converted_args
                arg = [arg]
            top_level_args.extend(
                a.array if isinstance(a, ScaledArrayView) else a for a in arg
            )

        args = converted_args
        ret = func(*args, **kwargs)
        if ret is not None:
            if isinstance(ret, np.ndarray) and ret.dtype != np.bool:
                return self.__class__(ret, self.scale, self.offset)
            if isinstance(ret, (bool, np.bool)):
                return ret
            else:
                return self._apply_scale(ret)
        return ret

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        inpts = []
        for inpt in inputs:
            if isinstance(inpt, self.__class__):
                inpts.append(inpt.array)
            else:
                inpts.append(inpt)
        ret = getattr(ufunc, method)(*inpts, **kwargs)
        if ret is not None:
            if isinstance(ret, np.ndarray):
                return self.__class__(ret, self.scale, self.offset)
            elif ret.dtype != np.bool:
                return self._apply_scale(ret)
            else:
                return ret
        return ret

    def __len__(self):
        return len(self.array)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.scale == other.scale
                and self.offset == other.offset
                and np.all(self.array == other.array)
            )
        else:
            return self.scaled_array() == other

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.scale != other.scale
                and self.offset != other.offset
                and np.all(self.array != other.array)
            )
        else:
            return self.scaled_array() != other

    def __lt__(self, other):
        return self.array < self._remove_scale(other)

    def __gt__(self, other):
        return self.array > self._remove_scale(other)

    def __ge__(self, other):
        return self.array >= self._remove_scale(other)

    def __le__(self, other):
        return self.array <= self._remove_scale(other)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._apply_scale(self.array[item])
        return self.__class__(self.array[item], self.scale, self.offset)

    def __setitem__(self, key, value):
        if isinstance(value, ScaledArrayView):
            iinfo = np.iinfo(self.array.dtype)
            if value.array.max() > iinfo.max or value.array.min() < iinfo.min:
                raise OverflowError(
                    "Values given do not fit after applying offest and scale"
                )
            self.array[key] = value.array[key]
        else:
            iinfo = np.iinfo(self.array.dtype)
            new_max = self._remove_scale(np.max(value))
            new_min = self._remove_scale(np.min(value))
            if new_max > iinfo.max or new_min < iinfo.min:
                raise OverflowError(
                    "Values given do not fit after applying offest and scale"
                )
            self.array[key] = self._remove_scale(value)

    def __repr__(self):
        return f"<ScaledArrayView({self.scaled_array()})>"



import abc
import io
import logging
import os
import subprocess
from typing import Optional, BinaryIO, Iterable, Union, Tuple

import numpy as np

from . import headers, errors, evlrs
from .compression import LazBackend
from .compression import find_laszip_executable
from .lasdatas import las14, las12
from .point import record, PointFormat
from .point.dims import size_of_point_format_id
from .typehints import LasData
from .utils import ConveyorThread
from .vlrs.known import LasZipVlr
from .vlrs.vlrlist import VLRList

try:
    import lazrs
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)


def get_extra_dims_info_tuple(header, vlrs) -> Optional[Tuple[Tuple[str, str], ...]]:
    try:
        extra_dims = vlrs.get("ExtraBytesVlr")[0].type_of_extra_dims()
    except IndexError:
        return None

    point_size_without_extra_bytes = size_of_point_format_id(header.point_format_id)
    if header.point_size == point_size_without_extra_bytes:
        logger.warning(
            "There is an ExtraByteVlr but the header.point_size matches the "
            "point size without extra bytes. The extra bytes vlr info will be ignored"
        )
        vlrs.extract("ExtraBytesVlr")
        extra_dims = None
    return extra_dims


class LasReader:
    """The reader class handles LAS and LAZ via one of the supported backend"""

    def __init__(
        self,
        source: BinaryIO,
        closefd: bool = True,
        laz_backend: Union[
            LazBackend, Iterable[LazBackend]
        ] = LazBackend.detect_available(),
    ):
        self.closefd = closefd
        self.laz_backend = laz_backend
        self.header, self.vlrs = self._read_header_and_vlrs(source)

        if self.header.are_points_compressed:
            if not laz_backend:
                raise errors.PylasError(
                    "No LazBackend selected, cannot decompress data"
                )
            self.point_source = self._create_laz_backend(source)
            if self.point_source is None:
                raise errors.PylasError(
                    "Data is compressed, but no LazBacked could be initialized"
                )
        else:
            self.point_source = UncompressedPointReader(source, self.header.point_size)

        self.points_read = 0
        self.point_format = PointFormat(
            self.header.point_format_id,
            extra_dims=get_extra_dims_info_tuple(self.header, self.vlrs),
        )

    def read_n_points(self, n: int) -> Optional[record.ScaleAwarePointRecord]:
        points_left = self.header.point_count - self.points_read
        if points_left <= 0:
            return None

        if n < 0:
            n = points_left
        else:
            n = min(n, points_left)

        r = record.PackedPointRecord.from_buffer(
            self.point_source.read_n_points(n), self.point_format, n
        )
        points = record.ScaleAwarePointRecord(
            r.array, r.point_format, self.header.scales, self.header.offsets
        )
        self.points_read += n
        return points

    def read(self) -> LasData:
        """Reads all the points not read and returns a LasData object"""
        points = self.read_n_points(-1)
        if points is None:
            points = record.PackedPointRecord.empty(self.point_format)

        if self.header.version >= "1.4":
            if (
                self.header.are_points_compressed
                and not self.point_source.source.seekable()
            ):
                # We explicitly require seekable stream because we have to seek
                # past the chunk table of LAZ file
                raise errors.PylasError(
                    "source must be seekable, to read evlrs form LAZ file"
                )
            evlrs = self._read_evlrs(self.point_source.source, seekable=True)
            las_data = las14.LasData(
                header=self.header, vlrs=self.vlrs, points=points, evlrs=evlrs
            )
        else:
            las_data = las12.LasData(header=self.header, vlrs=self.vlrs, points=points)

        return las_data

    def chunk_iterator(self, points_per_iteration: int) -> "PointChunkIterator":
        """Returns an iterator, that will read points by chunks
        of the requested size

        :param points_per_iteration: number of points to be read with each iteration
        :return:
        """
        return PointChunkIterator(self, points_per_iteration)

    def close(self) -> None:
        """closes the file object used by the reader"""
        if self.closefd:
            self.point_source.close()

    def _create_laz_backend(self, source) -> Optional["IPointReader"]:
        try:
            backends = iter(self.laz_backend)
        except TypeError:
            backends = (self.laz_backend,)

        laszip_vlr = self.vlrs.pop(self.vlrs.index("LasZipVlr"))
        for backend in backends:
            try:
                if not backend.is_available():
                    raise errors.PylasError(f"The '{backend}' is not available")

                if backend == LazBackend.LazrsParallel:
                    return LazrsPointReader(source, laszip_vlr, parallel=True)
                elif backend == LazBackend.Lazrs:
                    return LazrsPointReader(source, laszip_vlr, parallel=False)
                elif backend == LazBackend.Laszip:
                    point_source = LasZipProcessPointReader(
                        source, self.header, self.vlrs
                    )
                    self._read_header_and_vlrs(
                        point_source.process.stdout, seekable=False
                    )
                    return point_source
                else:
                    raise errors.PylasError("Unknown LazBackend: {}".format(backend))

            except errors.LazError as e:
                logger.error(e)

    @staticmethod
    def _read_header_and_vlrs(source, seekable=True):
        header = headers.HeaderFactory().read_from_stream(source)
        vlrs = VLRList.read_from(source, num_to_read=header.number_of_vlr)
        if seekable:
            offset = header.offset_to_point_data - source.tell()
            if offset >= 0:
                source.seek(offset, io.SEEK_CUR)
            else:
                raise RuntimeError("Read past point data")  # TODO
        return header, vlrs

    def _read_evlrs(self, source, seekable=False) -> Optional[evlrs.EVLRList]:
        """Reads the EVLRs of the file, will fail if the file version
        does not support evlrs
        """
        if self.header.version >= "1.4" and self.points_read == self.header.point_count:
            if seekable:
                source.seek(self.header.start_of_first_evlr)
            return evlrs.EVLRList.read_from(source, self.header.number_of_evlr)
        else:
            return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class PointChunkIterator:
    def __init__(self, reader, points_per_iteration) -> None:
        self.reader = reader
        self.points_per_iteration = points_per_iteration

    def __next__(self) -> record.ScaleAwarePointRecord:
        points = self.reader.read_n_points(self.points_per_iteration)
        if points is None:
            raise StopIteration
        return points

    def __iter__(self):
        return self


class IPointReader(abc.ABC):
    """The interface to be implemented by the class that actually reads
    points from as LAS/LAZ file so that the LasReader can use it.

    It is used to manipulate LAS/LAZ (with different LAZ backends) in the
    reader
    """

    @abc.abstractmethod
    def read_n_points(self, n) -> bytearray:
        ...

    @abc.abstractmethod
    def close(self):
        ...


class UncompressedPointReader(IPointReader):
    """Implementation of IPointReader for the simple uncompressed case"""

    def __init__(self, source, point_size) -> None:
        self.source = source
        self.point_size = point_size

    def read_n_points(self, n) -> bytearray:
        try:
            data = bytearray(n * self.point_size)
            self.source.readinto(data)
            return data
        except AttributeError:
            return bytearray(self.source.read(n * self.point_size))

    def close(self):
        self.source.close()


class LasZipProcessPointReader(IPointReader):
    """Implementation when using laszip executable as the LAZ backend.

    The compressed LAZ data (the whole file actually) is piped to laszip
    via its stdin and we get the uncompressed LAS data via its stdout


    when the source is a file object and not a file,
    we have to use a thread to move data from the file object to
    the laszip stdin to avoid a deadlock.
    """

    conveyor: Optional[ConveyorThread]

    def __init__(self, source, header, _vlrs) -> None:
        laszip_binary = find_laszip_executable()
        self.point_size: int = header.point_size
        if header.version >= "1.4" and header.number_of_evlr > 0:
            raise errors.PylasError(
                "Reading a LAZ file that contains EVLR using laszip is not supported"
            )
        try:
            fileno = source.fileno()
        except OSError:
            source.seek(0)
            self.source = source
            self.process = self.process = subprocess.Popen(
                [laszip_binary, "-stdin", "-olas", "-stdout"],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            self.conveyor = ConveyorThread(
                self.source, self.process.stdin, close_output=True
            )
            self.conveyor.start()
        else:
            os.lseek(fileno, 0, os.SEEK_SET)
            self.conveyor = None
            self.source = source
            self.process = self.process = subprocess.Popen(
                [laszip_binary, "-stdin", "-olas", "-stdout"],
                stdin=source,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

    def read_n_points(self, n) -> bytearray:
        b = bytearray(n * self.point_size)
        self.process.stdout.readinto(b)
        if self.process.poll() is not None:
            self.raise_if_bad_err_code()
        return bytearray(b)

    def close(self):
        if self.conveyor is not None and self.conveyor.is_alive():
            self.conveyor.ask_for_termination()
            self.conveyor.join()

        if self.process.poll() is None:
            # We are likely getting closed before decompressing all the data
            self.process.terminate()
            self.process.wait()
        else:
            self.raise_if_bad_err_code()

        self.process.stdout.close()
        self.process.stderr.close()
        self.source.close()

    def raise_if_bad_err_code(self):
        if self.process is not None and self.process.returncode != 0:
            error_msg = self.process.stderr.read().decode()
            raise RuntimeError(
                "Laszip failed to {} with error code {}:\n\t{}".format(
                    "compress",
                    self.process.returncode,
                    "\n\t".join(error_msg.splitlines()),
                )
            )


class LazrsPointReader(IPointReader):
    """Implementation for the laz-rs backend, supports single-threaded decompression
    as well as multi-threaded decompression
    """

    def __init__(self, source, laszip_vlr: LasZipVlr, parallel: bool) -> None:
        self.source = source
        self.vlr = lazrs.LazVlr(laszip_vlr.record_data)
        if parallel:
            self.decompressor = lazrs.ParLasZipDecompressor(
                source, laszip_vlr.record_data
            )
        else:
            self.decompressor = lazrs.LasZipDecompressor(source, laszip_vlr.record_data)

    def read_n_points(self, n) -> bytearray:
        point_bytes = bytearray(n * self.vlr.item_size())
        self.decompressor.decompress_many(point_bytes)
        return point_bytes

    def close(self):
        self.source.close()
